package se.lth.dit.odinmp;
import se.lth.dit.odinmp.*;
import se.lth.dit.odinmp.visitor.*;
import se.lth.dit.odinmp.syntaxtree.*;
import java.lang.*;
import java.util.*;

public class BarrierConstructDescriptor extends ConstructDescriptor {
    public static int nextIndex = 0;
    public static Vector inOrder = new Vector();

    public BarrierConstructDescriptor() {
	super(6, "barrier", nextIndex++);
	inOrder.addElement(this);
    }

    public String code(DataBuildingVisitor dbv) {
	Spacing spc = dbv.spc;
	StringBuffer retval = new StringBuffer();
	String otd = OdinMP.THREAD_INFO("odinmp_td");
	retval.append(nli(spc) + "{");
	spc.din();
	
	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    retval.append(nli(spc) + "double odinmp_timer_startTime, odinmp_timer_stopTime;");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append(nli(spc) + "if (odinmp_timing) {");
		spc.din();
	    }

	    retval.append(nli(spc) + "odinmp_timer_startTime = odinmp_time();");

	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		spc.dout();
		retval.append(nli(spc) + "}");
	    }

	    retval.append(nli(spc) + "{");
	    spc.din();
	}

	retval.append(nli(spc) + "odinmp_thread_data_t *odinmp_td = odinmp_self_data();");
	retval.append(nli(spc) + "if (" + otd + "->team != ((void *)0)) {");
	spc.din();
	retval.append(nli(spc) + "odinmp_barrier(" + otd + "->team, &" + otd + "->team->barriers[" + index + "]);");
	spc.dout();
	retval.append(nli(spc) + "}");
	
	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {

	    spc.dout();
	    retval.append("}");

	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append(nli(spc) + "if (odinmp_timing) {");
		spc.din();
	    }

	    retval.append(nli(spc) + "odinmp_timer_stopTime = odinmp_time();");
	    retval.append(nli(spc) + "odinmp_tacc_add(&odinmp_tacc_" + this + "_total, odinmp_timer_stopTime - odinmp_timer_startTime);");

	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		spc.dout();
		retval.append(nli(spc) + "}");
	    }
	}

	spc.dout();
	retval.append(nli(spc) + "}");
	return new String (retval);
    }


    public static String declareTiming() {
	StringBuffer retval = new StringBuffer();
	Enumeration e = inOrder.elements();
	while (e.hasMoreElements()) {
	    Object o = e.nextElement();
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_total;\n");
	}
	return new String(retval);
    }

    public static String externalDeclareTiming() {
	StringBuffer retval = new StringBuffer();
	Enumeration e = inOrder.elements();
	while (e.hasMoreElements()) {
	    Object o = e.nextElement();
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_total;\n");
	}
	return new String(retval);
    }

    public static String initTiming() {
	StringBuffer retval = new StringBuffer();
	Enumeration e = inOrder.elements();
	while (e.hasMoreElements()) {
	    Object o = e.nextElement();
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_total);\n");
	}
	return new String(retval);
    }

    public static String reportTiming() {
	Enumeration e = inOrder.elements();
	StringBuffer retval = new StringBuffer();
	while (e.hasMoreElements()) {
	    Object o = e.nextElement();
	    String s = o.toString();
	    
	    retval.append("   fprintf(stderr,\"Barrier " + s + ":\\n\");\n");
	    retval.append("   fprintf(stderr,\"  executed by a total of %d threads\\n\", odinmp_tacc_" + s + "_total.n);\n");
	    retval.append("   if (odinmp_tacc_" + s + "_total.n > 0) {\n");
	    retval.append("      fprintf(stderr,\"  total time:         min = %#12.7g s\\n\", odinmp_tacc_" + s + "_total.min);\n");
	    retval.append("      fprintf(stderr,\"                      avg = %#12.7g s\\n\", odinmp_tacc_" + s + "_total.total / odinmp_tacc_" + s + "_total.n);\n");
	    retval.append("      fprintf(stderr,\"                      max = %#12.7g s\\n\", odinmp_tacc_" + s + "_total.max);\n");
	    retval.append("   }\n");
	    retval.append("   fprintf(stderr,\"\\n\");\n");
	}
	return new String(retval);
    }

}
