package se.lth.dit.odinmp;
import se.lth.dit.odinmp.*;
import se.lth.dit.odinmp.syntaxtree.*;
import se.lth.dit.odinmp.visitor.*;
import java.lang.*;
import java.util.*;

public class SingleConstructDescriptor extends ConstructDescriptor {
    public static int nextIndex = 0;

    public static Vector inOrder = new Vector();

    public Statement statement;

    public SymbolTable privateSymbols = new SymbolTable();
    public SymbolTable firstprivateSymbols = new SymbolTable();

    public boolean nowait = false;

    public SingleConstructDescriptor(Statement cs) {
	super(4, "sing", nextIndex++);
	inOrder.addElement(this);
	statement = cs;
    }

    public String code(DataBuildingVisitor dbv) {
	Enumeration e;
	Symbol sym;
	SymbolTable declaredSymbols = new SymbolTable();
	OmpSymbolAttributes osa;
	String otd = OdinMP.THREAD_INFO("odinmp_td");
	
	StringBuffer retval = new StringBuffer();

	if (!dbv.lexicallyInParallel()) {
	    retval.append("  {\n");
	    retval.append("    odinmp_thread_data_t *odinmp_td = odinmp_self_data();\n");
	    retval.append("    if(((odinmp_thread_info_t *)odinmp_td)->team != ((void *)0)) {\n");
	    retval.append("\n");
	}

	retval.append("{\n");
	retval.append("  /* private variables; \n");
	retval.append("     if a variable is also firstprivate, then initialize it\n");
	retval.append("     from the corresponding outer variable */\n");

	/* Firstprivate variables */
	e = firstprivateSymbols.symbolsInOrder.elements();
	while (e.hasMoreElements()) {
	    sym = (Symbol)e.nextElement();
	    if (declaredSymbols.containsSymbol(sym))
		continue;
	    dbv.checkSharingForSymbol(sym);
	    String newName = "odinmp_flp_" + sym.name;
	    String nr = dbv.nameReplacementForSymbol(sym);
	    retval.append("  " + sym.nonStatic().typeString(newName) + " = " + nr + ";\n");
	    attributesForSymbol(sym).nameReplacement = newName;
	    declaredSymbols.addSymbol(sym);
	}
	
	/* "other" private variables */
	e = privateSymbols.symbolsInOrder.elements();
	while (e.hasMoreElements()) {
	    sym = (Symbol)e.nextElement();
	    if (declaredSymbols.containsSymbol(sym))
		continue;
	    osa = attributesForSymbol(sym);
	    String newName = sym.name;
	    retval.append("  " + sym + ";\n");
	    osa.nameReplacement = newName;
	    declaredSymbols.addSymbol(sym);
	}

	retval.append("  /* odinmp variables */\n");
	retval.append("  /* wsc info etc */\n");
	retval.append("  odinmp_sing_slice_t odinmp_slice;\n");
	retval.append("  odinmp_wsc_info_t odinmp_wsc_info;\n");
        retval.append("  int odinmp_ttid = " + otd + "->team_thread_id.id;\n");
	retval.append("\n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    retval.append("  double odinmp_timer_init_start, odinmp_timer_work_start, odinmp_timer_cleanup_start, odinmp_timer_barrier_start;\n");

	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  if (odinmp_timing) {\n");
	    }

	    retval.append("  odinmp_timer_init_start = odinmp_time();\n");

	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  }\n");
	    }
	    retval.append("  \n");
	}

	retval.append("  /* record info about this work sharing construct */\n");
	retval.append("  odinmp_wsc_info.next = " + otd + "->wsc_info;\n");
	retval.append("  " + otd + "->wsc_info = &odinmp_wsc_info;\n");
	retval.append("  odinmp_wsc_info.type = odinmp_wsc_type_sing;\n");
	retval.append("\n");
	retval.append("  /* fetch a slice */\n");
	retval.append("  {\n");
	retval.append("    odinmp_team_t *team = " + otd + "->team; \n");
	retval.append("    int lap;\n");
	retval.append("	  \n");
	retval.append("    ODINMP_TIMING_WAIT_BEGIN\n");
	retval.append("    pthread_mutex_lock(&team->lock);\n");
	retval.append("    ODINMP_TIMING_WAIT_END\n");
	retval.append("    lap = team->sing_laps[" + index + "];\n");
	retval.append("    if (!lap)\n");
	retval.append("      lap = team->thread_lap_infos[odinmp_ttid]." + this + ".lap;\n");
	retval.append("      team->sing_laps[" + index + "]++;\n");
	retval.append("	  \n");
	retval.append("    odinmp_slice.do_it = team->thread_lap_infos[odinmp_ttid]." + this + ".lap >= lap;\n");
	retval.append("    team->thread_lap_infos[odinmp_ttid]." + this + ".lap++;\n");
	retval.append("    pthread_mutex_unlock(&team->lock);\n");
	retval.append("  } /* fetch a slice */\n");
	retval.append("	\n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  if (odinmp_timing) {\n");
	    }
	    retval.append("  odinmp_tacc_add(&odinmp_tacc_" + this + "_init, (odinmp_timer_work_start = odinmp_time()) - odinmp_timer_init_start);\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  }\n");
	    }
	    retval.append("\n");
	}

	retval.append("  if (odinmp_slice.do_it) {\n");
	retval.append("    /* original code of single region comes here */\n");

	dbv.constructDescriptorStack.push(this);
	dbv.visit(statement);
	retval.append(dbv.rv.toString());
	dbv.constructDescriptorStack.pop();

	retval.append("  }\n");
	retval.append("\n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  if (odinmp_timing) {\n");
	    }
	    retval.append("  odinmp_tacc_add(&odinmp_tacc_" + this + "_work, (odinmp_timer_cleanup_start = odinmp_time()) - odinmp_timer_work_start);\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  }\n");
	    }
	    retval.append("\n");
	}

	retval.append("  /* restore work sharing construct info */\n");
	retval.append("  " + otd + "->wsc_info = odinmp_wsc_info.next;\n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  if (odinmp_timing) {\n");
	    }
	    retval.append("  odinmp_tacc_add(&odinmp_tacc_" + this + "_cleanup, odinmp_time() - odinmp_timer_work_start);\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  }\n");
	    }
	    retval.append("\n");
	}

	if (!nowait) {

	    if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
		if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		    retval.append("  if (odinmp_timing) {\n");
		}
		retval.append("  odinmp_timer_barrier_start = odinmp_time();\n");
		if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		    retval.append("  }\n");
		}
		retval.append("\n");
	    }

	    retval.append("  /* barrier at the end of single \n");
	    retval.append("     remove this section if nowait is specified */\n");
	    retval.append("  odinmp_barrier(" + otd + "->team, \n");
	    retval.append("		 &" + otd + "->team->sing_barriers[" + index + "]);\n");

	    if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
		if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		    retval.append("  if (odinmp_timing) {\n");
		}
		retval.append("  odinmp_tacc_add(&odinmp_tacc_" + this + "_barrier, odinmp_time() - odinmp_timer_barrier_start);\n");
		if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		    retval.append("  }\n");
		}
		retval.append("\n");
	    }

	}

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  if (odinmp_timing) {\n");
	    }
	    retval.append("  odinmp_tacc_add(&odinmp_tacc_" + this + "_total, odinmp_time() - odinmp_timer_init_start);\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  }\n");
	    }
	    retval.append("\n");
	}

	retval.append("  }\n");

	if (!dbv.lexicallyInParallel()) {
	    retval.append("    } else {\n");

	    dbv.visit(statement);
	    retval.append(dbv.rv.toString());

	    retval.append("    }\n");
	    retval.append("  }\n");
	    retval.append("\n");
	}

	return new String(retval);
    }

    public static String declareTiming() {
	StringBuffer retval = new StringBuffer();
	Enumeration e = inOrder.elements();
	while (e.hasMoreElements()) {
	    Object o = e.nextElement();
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_total;\n");
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_work;\n");
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_init;\n");
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_cleanup;\n");
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_barrier;\n");
	}
	return new String(retval);
    }

    public static String externalDeclareTiming() {
	StringBuffer retval = new StringBuffer();
	Enumeration e = inOrder.elements();
	while (e.hasMoreElements()) {
	    Object o = e.nextElement();
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_total;\n");
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_work;\n");
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_init;\n");
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_cleanup;\n");
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_barrier;\n");
	}
	return new String(retval);
    }

    public static String initTiming() {
	StringBuffer retval = new StringBuffer();
	Enumeration e = inOrder.elements();
	while (e.hasMoreElements()) {
	    Object o = e.nextElement();
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_total);\n");
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_work);\n");
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_init);\n");
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_cleanup);\n");
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_barrier);\n");
	}
	return new String(retval);
    }

    public static String reportTiming() {
	StringBuffer retval = new StringBuffer();
	Enumeration e = inOrder.elements();
	while (e.hasMoreElements()) {
	    Object o = e.nextElement();
	    String s = o.toString();
	    
	    retval.append("   fprintf(stderr,\"Single Construct " + s + ":\\n\");\n");
	    retval.append("   fprintf(stderr,\"  executed %d times total\\n\", odinmp_tacc_" + s + "_total.n);\n");
	    retval.append("   if (odinmp_tacc_" + s + "_total.n > 0) {\n");
	    retval.append("      fprintf(stderr,\"  total of %d iterations of work, %5f iterations/execution on average\\n\", odinmp_tacc_" + s + "_work.n, (double)odinmp_tacc_" + s + "_work.n / (double)odinmp_tacc_" + s + "_total.n);\n");
	    retval.append("      fprintf(stderr,\"  total time:         min = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_total.min);\n");
	    retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_total.total / odinmp_tacc_" + s + "_total.n);\n");
	    retval.append("      fprintf(stderr,\"                      max = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_total.max);\n");
	    retval.append("      fprintf(stderr,\"                    total = %#12.7g s accumulated\\n\", odinmp_tacc_" + s + "_total.total);\n");

	    retval.append("      fprintf(stderr,\"  working time:       min = %#12.7g s per iteration per execution\\n\", odinmp_tacc_" + s + "_work.min);\n");
	    retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per iteration per execution\\n\", odinmp_tacc_" + s + "_work.total / odinmp_tacc_" + s + "_work.n);\n");
	    retval.append("      fprintf(stderr,\"                      max = %#12.7g s per iteration per execution\\n\", odinmp_tacc_" + s + "_work.max);\n");
	    retval.append("      fprintf(stderr,\"                    total = %#12.7g s accumulated\\n\", odinmp_tacc_" + s + "_work.total);\n");

	    retval.append("      fprintf(stderr,\"  initialization:     min = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_init.min);\n");
	    retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_init.total / odinmp_tacc_" + s + "_init.n);\n");
	    retval.append("      fprintf(stderr,\"                      max = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_init.max);\n");
	    retval.append("      fprintf(stderr,\"                    total = %#12.7g s accumulated\\n\", odinmp_tacc_" + s + "_init.total);\n");

	    retval.append("      fprintf(stderr,\"  cleanup:            min = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_cleanup.min);\n");
	    retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_cleanup.total / odinmp_tacc_" + s + "_cleanup.n);\n");
	    retval.append("      fprintf(stderr,\"                      max = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_cleanup.max);\n");
	    retval.append("      fprintf(stderr,\"                    total = %#12.7g s accumulated\\n\", odinmp_tacc_" + s + "_cleanup.total);\n");

	    if (!((SingleConstructDescriptor)o).nowait) {
		retval.append("      fprintf(stderr,\"  barrier:            min = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_barrier.min);\n");
		retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_barrier.total / odinmp_tacc_" + s + "_cleanup.n);\n");
		retval.append("      fprintf(stderr,\"                      max = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_barrier.max);\n");
		retval.append("      fprintf(stderr,\"                    total = %#12.7g s accumulated\\n\", odinmp_tacc_" + s + "_barrier.total);\n");
	    } else {
		retval.append("      fprintf(stderr,\"  " + o + " has nowait specified, so no barrier at the end\\n\");\n");
	    }

	    retval.append("   }\n");
	    retval.append("   fprintf(stderr,\"\\n\");\n");
	}
	return new String(retval);
    }
}
