package se.lth.dit.odinmp;
import se.lth.dit.odinmp.*;
import se.lth.dit.odinmp.syntaxtree.*;
import se.lth.dit.odinmp.visitor.*;
import java.util.*;

public class ForClauseHandler extends SymbolClauseHandler {
    public ForConstructDescriptor forConstructDescriptor;

    public ForClauseHandler(SymbolFetcher sf, ForConstructDescriptor forCd) {
	super(sf);
	forConstructDescriptor = forCd;
    }

    public void handlePrivate(Vector v) {
	Vector syms = symbolsForNames(v);
	setSharingForSymbolsInConstruct(OmpSymbolAttributes.sharing_Private,
					syms,
					forConstructDescriptor);
	forConstructDescriptor.privateSymbols.addSymbols(syms);
    }

    public void handleFirstprivate(Vector v) {
	Vector syms = symbolsForNames(v);
	setSharingForSymbolsInConstruct(OmpSymbolAttributes.sharing_Firstprivate,
					syms,
					forConstructDescriptor);
	forConstructDescriptor.firstprivateSymbols.addSymbols(syms);
    }

    public void handleLastprivate(Vector v) {
	Vector syms = symbolsForNames(v);
	setSharingForSymbolsInConstruct(OmpSymbolAttributes.sharing_Lastprivate,
					syms,
					forConstructDescriptor);
	forConstructDescriptor.lastprivateSymbols.addSymbols(syms);
    }
    
    /*
      public void handleShared(Vector v) {
      }
    */

    /*
      public void handleCopyin(Vector v) {
      }
    */

    /*
      public void handleDefault(String s) {
      }
    */

    public void handleReduction(String redop, Vector v) {
	Vector vv = symbolsForNames(v);
	Enumeration e = vv.elements();
	while (e.hasMoreElements()) {
	    Symbol s = (Symbol)e.nextElement();
	    OmpSymbolAttributes osa = forConstructDescriptor.attributesForSymbol(s);
	    osa.setSharing(OmpSymbolAttributes.sharing_Reduction);
	    osa.reductionOperator = redop;
	    forConstructDescriptor.reductionSymbols.addSymbol(s);

	    /* OdinMP.debugPrintln("reduction clause: osa = " + osa + ", symbol ``" + s + "'', red_op ``" + osa.reductionOperator + "''"); */
	    
	}
    }
    
    /*
      public void handleIf(String condition) {
      }
    */
    
    public void handleNowait() {
	forConstructDescriptor.nowait = true;
    }
    
    public void handleOrdered() {
	forConstructDescriptor.ordered = true;
    }
    
    public void handleSchedule(String scheduleKind, Expression chunkSize) {
	if (scheduleKind.equals("static")) {
	    forConstructDescriptor.scheduler = ForConstructDescriptor.scheduler_Static;
	} else if (scheduleKind.equals("dynamic")) {
	    forConstructDescriptor.scheduler = ForConstructDescriptor.scheduler_Dynamic;
	} else if (scheduleKind.equals("guided")) {
	    forConstructDescriptor.scheduler = ForConstructDescriptor.scheduler_Guided;
	} else if (scheduleKind.equals("runtime")) {
	    forConstructDescriptor.scheduler = ForConstructDescriptor.scheduler_Runtime;
	} else {
	    // huh ? should never happen
	}
	forConstructDescriptor.chunkSize = chunkSize;
    }
}
