package se.lth.dit.odinmp;
import se.lth.dit.odinmp.*;
import se.lth.dit.odinmp.syntaxtree.*;
import se.lth.dit.odinmp.visitor.*;
import java.lang.*;
import java.util.*;

public class ForConstructDescriptor extends ConstructDescriptor {
    public static int nextIndex = 0;
    
    public static Vector inOrder = new Vector();

    public static final int comparisonOperator_LT = -2;
    public static final int comparisonOperator_LE = -1;
    public static final int comparisonOperator_GE =  1;
    public static final int comparisonOperator_GT =  2;

    public static final int scheduler_Static = 0;
    public static final int scheduler_Dynamic = 1;
    public static final int scheduler_Guided = 2;
    public static final int scheduler_Runtime = 3;

    public static final String[] schedulerType = {
        "odinmp_scheduler_static",
        "odinmp_scheduler_dynamic",
        "odinmp_scheduler_guided",
        "odinmp_scheduler_runtime"
    };

    public SymbolTable privateSymbols = new SymbolTable();
    public SymbolTable reductionSymbols = new SymbolTable();
    public SymbolTable firstprivateSymbols = new SymbolTable();
    public SymbolTable lastprivateSymbols = new SymbolTable();
    public SourceFile sourceFile;

    public boolean ordered = false;
    public boolean nowait = false;

    public Symbol loopIndex;
    public Expression initExpression;
    public int comparisonOperator;
    public Expression compareAgainst;

    public int incr_mult;
    public Expression incr_Expression;
    public AdditiveExpression incr_AdditiveExpression;
    public MultiplicativeExpression incr_MultiplicativeExpression;

    public int scheduler = -1;
    public Expression chunkSize;

    public Statement statement = null;

    public ForConstructDescriptor() {
        super(2, "for", nextIndex++);
        inOrder.addElement(this);
    }

    public ForConstructDescriptor(Statement s) {
        this();
        statement = s;
    }

    public void setIncrement(int mul, Expression expr) {
        incr_mult = mul;
        incr_Expression = expr;
        incr_AdditiveExpression = null;
        incr_MultiplicativeExpression = null;
    }

    public void setIncrement(int incr) {
        incr_mult = incr;
        incr_Expression = null;
        incr_AdditiveExpression = null;
        incr_MultiplicativeExpression = null;
    }

    public void setIncrement(int mul, AdditiveExpression expr) {
        incr_mult = mul;
        incr_Expression = null;
        incr_AdditiveExpression = expr;
        incr_MultiplicativeExpression = null;
    }

    public void setIncrement(int mul, MultiplicativeExpression expr) {
        incr_mult = mul;
        incr_Expression = null;
        incr_AdditiveExpression = null;
        incr_MultiplicativeExpression = expr;
    }

    public String increment(DataBuildingVisitor dbv) {
        if (incr_Expression != null) {
            dbv.visit(incr_Expression);
            if (incr_mult < 0) 
                return "-(" + dbv.rv.toString() + ")";
            return dbv.rv.toString();
        } else if (incr_AdditiveExpression != null) {
            dbv.visit(incr_AdditiveExpression);
            if (incr_mult < 0) 
                return "-(" + dbv.rv.toString() + ")";
            return dbv.rv.toString();
        } else if (incr_MultiplicativeExpression != null) {
            dbv.visit(incr_MultiplicativeExpression);
            if (incr_mult < 0) 
                return "-(" + dbv.rv.toString() + ")";
            return dbv.rv.toString();
        } else { // ++ or --
            if (incr_mult < 0) 
                return "-1";
            return "1";
        }
    }

    public String lowerBound(DataBuildingVisitor dbv) {
        dbv.visit(initExpression);
        return dbv.rv.toString();
    }

    public String bound(DataBuildingVisitor dbv) {
        dbv.visit(compareAgainst);
        return dbv.rv.toString();
    }

    public String code(DataBuildingVisitor dbv) {
        Enumeration e;
        Symbol sym;
        SymbolTable declaredSymbols = new SymbolTable();
        OmpSymbolAttributes osa;

        String otd = OdinMP.THREAD_INFO("odinmp_td");
        String owf = "odinmp_wsc_info.u.wsc_for";
        String oll = "odinmp_wsc_info.u.wsc_for.lap_list";
        String incr;
        String initExtra = "";
        String compOp = "==";

        switch (comparisonOperator) {
        case comparisonOperator_LE:
            initExtra = " + 1";
            compOp = "<";
            break;
        case comparisonOperator_LT:
            initExtra = "";
            compOp = "<";
            break;
        case comparisonOperator_GE:
            initExtra = " - 1";
            compOp = ">";
            break;
        case comparisonOperator_GT:
            initExtra = "";
            compOp = ">";
            break;
        default:
            // huh ? not supposed to happen 
        }

        sourceFile = dbv.odinmp.currentSourceFile;

        /* OdinMP.debugPrintln(this + ": generating code"); */

        StringBuffer retval = new StringBuffer();

        if (!dbv.lexicallyInParallel()) {
            retval.append("  {\n");
            retval.append("    odinmp_thread_data_t *odinmp_td = odinmp_self_data();\n");
            retval.append("    if(((odinmp_thread_info_t *)odinmp_td)->team != ((void *)0)) {\n");
            retval.append("\n");
        }

        if (scheduler < 0) 
            scheduler = scheduler_Static;

        retval.append("{\n");
        retval.append("  /* private variables; \n");
        retval.append("     if a variable is also firstprivate, then initialize it\n");
        retval.append("     from the corresponding outer variable\n");
        retval.append("     if a variable is a reduction variable, then initialize it\n");
        retval.append("     corresponding to the reduction operator */\n");

        /* Add the Loop Index to private variables */
        if (!privateSymbols.containsSymbol(loopIndex)) {
            attributesForSymbol(loopIndex).setSharing(OmpSymbolAttributes.sharing_Private);
            privateSymbols.addSymbol(loopIndex);
        }

        /* Firstprivate variables */
        e = firstprivateSymbols.symbolsInOrder.elements();
        while (e.hasMoreElements()) {
            sym = (Symbol)e.nextElement();
            if (declaredSymbols.containsSymbol(sym))
                continue;
            dbv.checkSharingForSymbol(sym);
            String newName = "odinmp_flp_" + sym.name;
            String nr = dbv.nameReplacementForSymbol(sym);
            retval.append("  " + sym.nonStatic().typeString(newName) + " = " + nr + ";\n");
            attributesForSymbol(sym).nameReplacement = newName;
            declaredSymbols.addSymbol(sym);
        }
        
        /* Lastprivate variables */
        e = lastprivateSymbols.symbolsInOrder.elements();
        while (e.hasMoreElements()) {
            sym = (Symbol)e.nextElement();
            if (declaredSymbols.containsSymbol(sym))
                continue;
            dbv.checkSharingForSymbol(sym);
            String newName = "odinmp_flp_" + sym.name;
            retval.append("  " + sym.nonStatic().typeString(newName) + ";\n");
            attributesForSymbol(sym).nameReplacement = newName;
            declaredSymbols.addSymbol(sym);
        }

        /* Reduction variables */
        e = reductionSymbols.symbolsInOrder.elements();
        while (e.hasMoreElements()) {
            sym = (Symbol)e.nextElement();
            if (declaredSymbols.containsSymbol(sym))
                continue;
            dbv.checkSharingForSymbol(sym);
            osa = attributesForSymbol(sym);
            String newName = "odinmp_rdx_" + sym.name;
            retval.append("  " + sym.nonStatic().typeString(newName) + " = " + osa.reductionInitValue() + ";\n");
            osa.nameReplacement = newName;
            declaredSymbols.addSymbol(sym);
        }

        /* "other" private variables */
        e = privateSymbols.symbolsInOrder.elements();
        while (e.hasMoreElements()) {
            sym = (Symbol)e.nextElement();
            if (declaredSymbols.containsSymbol(sym))
                continue;
            osa = attributesForSymbol(sym);
            String newName = sym.name;
            retval.append("  " + sym + ";\n");
            osa.nameReplacement = newName;
            declaredSymbols.addSymbol(sym);
        }
        
        retval.append("  \n");
        retval.append("  /* odinmp variables */\n");
        retval.append("  /* wsc info etc */\n");
        retval.append("  odinmp_for_slice_t odinmp_slice = { 0L, 0L, 0, 0, 0, 0 };\n");
        retval.append("  odinmp_wsc_info_t odinmp_wsc_info;\n");
        if (ordered)
            retval.append("  int odinmp_in_order = 0;\n");
        retval.append("  /* for slice fetching */\n");
        retval.append("  odinmp_team_t *odinmp_team = " + otd + "->team;\n");
        retval.append("  pthread_mutex_t *odinmp_team_lock = &odinmp_team->lock;\n");
        retval.append("  int odinmp_ttid = " + otd + "->team_thread_id.id;\n");
        retval.append("  \n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    retval.append("  double odinmp_timer_init_start, odinmp_timer_init_iter_start, odinmp_timer_work_start, odinmp_timer_cleanup_iter_start, odinmp_timer_cleanup_start, odinmp_timer_barrier_start;\n");

	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  if (odinmp_timing) {\n");
	    }

	    retval.append("  odinmp_timer_init_start = odinmp_time();\n");

	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  }\n");
	    }
	    retval.append("  \n");
	}

        retval.append("  /* record info about this work sharing construct */\n");
        retval.append("  odinmp_wsc_info.next = " + otd + "->wsc_info;\n");
        retval.append("  " + otd + "->wsc_info = &odinmp_wsc_info;\n");
        retval.append("  odinmp_wsc_info.type = odinmp_wsc_type_for;\n");
        retval.append("  " + owf + ".slice = &odinmp_slice;\n");
        retval.append("\n");
        retval.append("  ODINMP_TIMING_WAIT_BEGIN\n");
        retval.append("  pthread_mutex_lock(odinmp_team_lock);\n");
        retval.append("  ODINMP_TIMING_WAIT_END\n");
        retval.append("      \n");
        retval.append("  /* find the lap list, if any */\n");
        retval.append("  for (" + oll + " = odinmp_team->for_laps[" + index + "];\n");
        retval.append("       " + oll + " && " + oll + "->lap > odinmp_team->thread_lap_infos[odinmp_ttid]." + this + ".lap;\n");
        retval.append("       " + oll + " = " + oll + "->next)\n");
        retval.append("      ;\n");
        retval.append("\n");
        retval.append("  if (!" + oll + " || " + oll + "->lap != odinmp_team->thread_lap_infos[odinmp_ttid]." + this + ".lap) {\n");
        retval.append("      " + OdinMP.PRINTF("\"thread %d: beginning new lap (%d) of " + this + "\\n\", " + otd + "->id, odinmp_team->thread_lap_infos[odinmp_ttid]." + this + ".lap") + ";\n");
        retval.append("      \n");
        retval.append("      " + oll + " = (odinmp_for_lap_list_t *)\n");
        retval.append("         malloc(sizeof(odinmp_for_lap_list_t));\n");
        retval.append("      " + oll + "->lap = odinmp_team->thread_lap_infos[odinmp_ttid]." + this + ".lap;\n");
        retval.append("      " + oll + "->n_left_to_leave = odinmp_team->n;\n");
        retval.append("      " + oll + "->for_data.lb = " + lowerBound(dbv) + ";\n");
        retval.append("      " + oll + "->for_data.incr = " + increment(dbv) + ";\n");
        retval.append("      " + oll + "->for_data.b = (" + bound(dbv) + ")" + initExtra + ";\n");
        retval.append("      " + oll + "->for_data.scheduler.type = " + schedulerType[scheduler] + ";\n");
        retval.append("      " + oll + "->for_data.scheduler.chunk_size = ");
        
        if (chunkSize == null) {
            if (scheduler == scheduler_Static) {
                retval.append("(" + oll + "->for_data.b - " + oll + "->for_data.lb) / (double)odinmp_team->n");
            } else if (scheduler == scheduler_Dynamic || scheduler == scheduler_Dynamic) {
                retval.append("1");
            } else if (scheduler == scheduler_Runtime) {
                // runtime handling
		retval.append("(odinmp_runtime_scheduler.chunk_size == 0.0) ? ((odinmp_runtime_scheduler.type == odinmp_scheduler_static) ? ((" + oll + "->for_data.b - " + oll + "->for_data.lb) / (double)odinmp_team->n) : 1) : odinmp_runtime_scheduler.chunk_size");
            } else {
                // huh ? should never happen
            }
        } else {
            chunkSize.accept(dbv);
            retval.append(dbv.rv.toString());
        }
        
        retval.append(";\n");

        if (scheduler == scheduler_Static) {
            retval.append("      " + oll + "->for_data.chunk_size = " + oll + "->for_data.scheduler.chunk_size;\n");
        } else {
            retval.append("      " + oll + "->for_data.chunk_size = " + oll + "->for_data.scheduler.chunk_size * " + oll + "->for_data.incr;\n");
        }

        retval.append("      " + OdinMP.PRINTF("\"for_data->chunk_size = %lf\\n\", " + oll + "->for_data.chunk_size") + ";\n");
        retval.append("      " + OdinMP.PRINTF("\"for_data->scheduler.chunk_size = %lf\\n\", " + oll + "->for_data.scheduler.chunk_size") + ";\n");

        retval.append("      " + oll + "->next = odinmp_team->for_laps[" + index + "];\n");
        retval.append("      odinmp_team->for_laps[" + index + "] = " + oll + ";\n");
        retval.append("      \n");
        
        if (ordered) {
            retval.append("      /* if ordered clause is specified on for construct */\n");
            retval.append("      " + oll + "->for_data.in_order = 0;\n");
            retval.append("      " + oll + "->for_data.next_in_order = 0;\n");
            retval.append("      pthread_mutex_init(&" + oll + "->for_data.order_lock, ((void *)0));\n");
            retval.append("      pthread_cond_init(&" + oll + "->for_data.new_order, ((void *)0));\n");
        }
        
        retval.append("  }\n");
        retval.append("\n");
        retval.append("  pthread_mutex_unlock(odinmp_team_lock);\n");
        retval.append("  \n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  if (odinmp_timing) {\n");
	    }
	    retval.append("  odinmp_tacc_add(&odinmp_tacc_" + this + "_init, odinmp_time() - odinmp_timer_init_start);\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  }\n");
	    }
	    retval.append("  \n");
	}
	
        retval.append("  while (1) {\n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  if (odinmp_timing) {\n");
	    }
	    retval.append("  odinmp_timer_init_iter_start = odinmp_time();\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  }\n");
	    }
	    retval.append("  \n");
	}

        retval.append("    /* fetch a slice */");
        retval.append("\n");

        retval.append("    " + OdinMP.PRINTF("\"thread %d: preparing to get slice for " + this + "\\n\", " + otd + "->id") + ";\n");
        retval.append("\n");
        retval.append("    /* reset slice */\n");
        retval.append("    odinmp_slice.last = 0;\n");
        retval.append("    odinmp_slice.done = 0;\n");
        retval.append("    odinmp_slice.lb = 0;\n");
        retval.append("    odinmp_slice.incr = " + oll + "->for_data.incr;\n");
        retval.append("    odinmp_slice.b = 0;\n");
        retval.append("    \n");
        retval.append("    " + OdinMP.PRINTF("\"thread %d, " + this + " remaining: lb = %d, incr = %d, b = %d\\n\", " + otd + "->id, (int)" + oll + "->for_data.lb, (int)" + oll + "->for_data.incr, (int)" + oll + "->for_data.b") + ";\n");
        retval.append("    \n");
        
        if (scheduler == scheduler_Runtime) {
            retval.append("    switch(odinmp_runtime_scheduler.type) {\n");
            retval.append("    case odinmp_scheduler_static:\n");
            retval.append("\n");
        }
        
        if (scheduler == scheduler_Runtime || scheduler == scheduler_Static) {
            retval.append("      {\n");
            retval.append("        int i = " + otd + "->id \n");
            retval.append("                + odinmp_slice.i * odinmp_team->n;\n");
            retval.append("\n");
            retval.append("        odinmp_slice.lb = \n");
            retval.append("          (long long int)(" + oll + "->for_data.lb + i * " + oll + "->for_data.chunk_size);\n");
            retval.append("\n");
            retval.append("        if (odinmp_slice.lb " + compOp + " " + oll + "->for_data.b) {\n");
            retval.append("          odinmp_slice.b = \n");
            retval.append("            (long long int)(" + oll + "->for_data.lb + (i + 1) * " + oll + "->for_data.chunk_size);\n");
            retval.append("          if (!(odinmp_slice.b " + compOp + " " + oll + "->for_data.b)) {\n");
            retval.append("            odinmp_slice.b = " + oll + "->for_data.b;\n");
            retval.append("            odinmp_slice.last = 1;\n");
            retval.append("          } \n");

            if (ordered) {
                retval.append("          odinmp_slice.in_order = i;\n");
                retval.append("\n");
            }

            retval.append("          odinmp_slice.i++;\n");
            retval.append("        } else {\n");
            retval.append("          odinmp_slice.done = 1;\n");
            retval.append("        }\n");
            retval.append("      }\n");
        }

        if (scheduler == scheduler_Runtime) {
            retval.append("      break;\n");
            retval.append("    case odinmp_scheduler_dynamic:\n");
            retval.append("\n");
        }

        if (scheduler == scheduler_Runtime || scheduler == scheduler_Dynamic) {
	    retval.append("      ODINMP_TIMING_WAIT_BEGIN\n");
            retval.append("      pthread_mutex_lock(odinmp_team_lock);\n");
	    retval.append("      ODINMP_TIMING_WAIT_END\n");
            retval.append("      if (" + oll + "->for_data.lb " + compOp + " " + oll + "->for_data.b) {\n");
            retval.append("        odinmp_slice.lb = " + oll + "->for_data.lb;\n");
            retval.append("        " + oll + "->for_data.lb = \n");
            retval.append("          (long long int)(" + oll + "->for_data.lb + " + oll + "->for_data.chunk_size);\n");
            retval.append("        odinmp_slice.b = " + oll + "->for_data.lb;\n");
            retval.append("\n");
            retval.append("        if (!(" + oll + "->for_data.lb " + compOp + " " + oll + "->for_data.b)) {\n");
            retval.append("          odinmp_slice.b = " + oll + "->for_data.b;\n");
            retval.append("          odinmp_slice.last = 1;\n");
            retval.append("        } \n");

            if (ordered) {
                retval.append("        odinmp_slice.in_order = " + oll + "->for_data.next_in_order++;\n");
            }

            retval.append("      } else {\n");
            retval.append("        odinmp_slice.lb = 0;\n");
            retval.append("        odinmp_slice.b = -1;\n");
            retval.append("        odinmp_slice.last = 0;\n");
            retval.append("        odinmp_slice.done = 1;\n");
            retval.append("      }\n");
            retval.append("      pthread_mutex_unlock(odinmp_team_lock);\n");
        }

        if (scheduler == scheduler_Runtime) {
            retval.append("    break;\n");
            retval.append("      case odinmp_scheduler_guided:\n");
        }

        if (scheduler == scheduler_Runtime || scheduler == scheduler_Guided) {
	    retval.append("      ODINMP_TIMING_WAIT_BEGIN\n");
            retval.append("      pthread_mutex_lock(odinmp_team_lock);\n");
	    retval.append("      ODINMP_TIMING_WAIT_END\n");
            retval.append("      if (" + oll + "->for_data.lb " + compOp + " " + oll + "->for_data.b) {\n");
            retval.append("        long long int space_left = (" + oll + "->for_data.b - " + oll + "->for_data.lb);\n");
            retval.append("        if ((long long int)(space_left / " + oll + "->for_data.incr) > " + oll + "->for_data.scheduler.chunk_size) {\n");
            retval.append("          odinmp_slice.lb = " + oll + "->for_data.lb;\n");
            retval.append("          " + oll + "->for_data.lb = \n");
            retval.append("            (long long int)(" + oll + "->for_data.lb + (((space_left - " + oll + "->for_data.chunk_size) / odinmp_team->n) + " + oll + "->for_data.chunk_size));\n");
            retval.append("          odinmp_slice.b = " + oll + "->for_data.lb;\n");
            retval.append("        } else {\n");
            retval.append("          odinmp_slice.lb = " + oll + "->for_data.lb;\n");
            retval.append("          odinmp_slice.b = " + oll + "->for_data.b;\n");
            retval.append("          " + oll + "->for_data.lb = " + oll + "->for_data.b;\n");
            retval.append("          odinmp_slice.last = 1;\n");
            retval.append("        }\n");

            if (ordered) {
                retval.append("        odinmp_slice.in_order = " + oll + "->for_data.next_in_order++;\n");
            }

            retval.append("      } else {\n");
            retval.append("        odinmp_slice.lb = 0;\n");
            retval.append("        odinmp_slice.b = -1;\n");
            retval.append("        odinmp_slice.last = 0;\n");
            retval.append("        odinmp_slice.done = 1;\n");
            retval.append("      }\n");
            retval.append("      pthread_mutex_unlock(odinmp_team_lock);\n");
        }

        if (scheduler == scheduler_Runtime) {
            retval.append("      break;\n");
            retval.append("    }\n");
        }


	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  if (odinmp_timing) {\n");
	    }
	    retval.append("  odinmp_tacc_add(&odinmp_tacc_" + this + "_init_iter, odinmp_time() - odinmp_timer_init_iter_start);\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  }\n");
	    }
	    retval.append("  \n");
	}

        retval.append("\n");
        retval.append("    " + OdinMP.PRINTF("\"thread %d: got slice for " + this + ": lap = %d, lb = %d, b = %d, last = %d, done = %d\\n\"," + otd + "->id, (int)odinmp_team->thread_lap_infos[odinmp_ttid]." + this + ".lap, (int)odinmp_slice.lb, (int)odinmp_slice.b, (int)odinmp_slice.last, (int)odinmp_slice.done") + ";\n");
        retval.append("\n");

        retval.append("    if (odinmp_slice.done) {\n");
        retval.append("      /* increment lap counter */\n");
        retval.append("      odinmp_team->thread_lap_infos[odinmp_ttid]." + this + ".lap ++;\n");
        retval.append("      /* ... and exit the while(1) loop */\n");
        retval.append("      break;\n");
        retval.append("    }\n");
        retval.append("\n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  if (odinmp_timing) {\n");
	    }
	    retval.append("  odinmp_timer_work_start = odinmp_time();\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  }\n");
	    }
	    retval.append("\n");
	}

        retval.append("    /* here's the original loop, with the for loop head exchanged \n");
        retval.append("       and variable references replaced as well */\n");
        retval.append("    for (" + attributesForSymbol(loopIndex).nameReplacement + " = odinmp_slice.lb;\n");
        retval.append("         " + attributesForSymbol(loopIndex).nameReplacement + " " + compOp + " odinmp_slice.b;\n");
        retval.append("         " + attributesForSymbol(loopIndex).nameReplacement + " += odinmp_slice.incr)\n");

        dbv.constructDescriptorStack.push(this);

        if (statement != null) {
            dbv.visit(statement);
            retval.append(dbv.rv.toString());
        } else {
            // huh ? should never happen
            retval.append("/* No generated code, should never happen */\n");
        }

        dbv.constructDescriptorStack.pop();

        retval.append("    \n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  if (odinmp_timing) {\n");
	    }
	    retval.append("    odinmp_tacc_add(&odinmp_tacc_" + this + "_work, (odinmp_timer_cleanup_iter_start = odinmp_time()) - odinmp_timer_work_start);\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  }\n");
	    }
	}

        if (ordered) {
	    retval.append("    ODINMP_TIMING_WAIT_BEGIN\n");
            retval.append("    pthread_mutex_lock(&" + oll + "->for_data.order_lock);\n");
	    retval.append("    ODINMP_TIMING_WAIT_END\n");
            retval.append("    " + oll + "->for_data.in_order++;\n");
            retval.append("    pthread_cond_broadcast(&" + oll + "->for_data.new_order);\n");
            retval.append("    pthread_mutex_unlock(&" + oll + "->for_data.order_lock);\n");
            retval.append("\n");
        }

        retval.append("    /* did we run the last iteration ? */\n");
        retval.append("    if (odinmp_slice.last) {\n");
        retval.append("      /* yes, so copy data out from lastprivate variables */\n");

        /* copy data out to lastprivate variables */
        e = lastprivateSymbols.symbolsInOrder.elements();
        while (e.hasMoreElements()) {
            sym = (Symbol)e.nextElement();
            osa = attributesForSymbol(sym);
            String newName = osa.nameReplacement;
            osa.nameReplacement = null;
            retval.append("    " + dbv.nameReplacementForSymbol(sym) + " = " + newName + ";\n");
        }

        retval.append("    }\n");
        retval.append("    \n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  if (odinmp_timing) {\n");
	    }
	    retval.append("    odinmp_tacc_add(&odinmp_tacc_" + this + "_cleanup_iter, odinmp_time() - odinmp_timer_cleanup_iter_start);\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  }\n");
	    }
	    retval.append("  \n");
	}

        retval.append("  } /* while(1) */\n");
        retval.append("\n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  if (odinmp_timing) {\n");
	    }
	    retval.append("  odinmp_timer_cleanup_start = odinmp_time();\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  }\n");
	    }
	    retval.append("  \n");
	}

        retval.append("  /* restore work sharing construct info */\n");
        retval.append("  " + otd + "->wsc_info = odinmp_wsc_info.next;\n");
        retval.append("  \n");

	retval.append("  ODINMP_TIMING_WAIT_BEGIN\n");
        retval.append("  pthread_mutex_lock(odinmp_team_lock);\n");
	retval.append("  ODINMP_TIMING_WAIT_END\n");
        retval.append("  \n");
        retval.append("  /* reduce reduction variables */\n");
        /* reduce reduction variables */
        e = reductionSymbols.symbolsInOrder.elements();
        if (e.hasMoreElements()) {
            while (e.hasMoreElements()) {
                sym = (Symbol)e.nextElement();
                osa = attributesForSymbol(sym);
                String newName = osa.nameReplacement;
                osa.nameReplacement = null;
                String nr = dbv.nameReplacementForSymbol(sym);
                retval.append("  " + nr + " = " + nr + " " + osa.reductionOperator + " " + newName + ";\n");
            }
        }
        
        retval.append("  /* update the lap list */\n");
        retval.append("  if (--" + oll + "->n_left_to_leave == 0) {\n");
        retval.append("    odinmp_for_lap_list_t *oll = odinmp_team->for_laps[" + index + "];\n");
        retval.append("    " + OdinMP.PRINTF("\"thread %d: ending lap (%d) of " + this + "\\n\", " + otd + "->id, odinmp_team->thread_lap_infos[odinmp_ttid]." + this + ".lap") + ";\n");
        retval.append("    if (oll == " + oll + ") {\n");
        retval.append("      odinmp_team->for_laps[" + index + "] = oll->next;\n");
        retval.append("    } else {\n");
        retval.append("      while (oll->next != " + oll + ")\n");
        retval.append("        oll = oll->next;\n");
        retval.append("      oll->next = " + oll + "->next;\n");
        retval.append("    }\n");
        retval.append("    free(oll);\n");
        retval.append("  }\n");
        retval.append("  pthread_mutex_unlock(odinmp_team_lock);\n");
        
	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  if (odinmp_timing) {\n");
	    }
	    retval.append("  odinmp_tacc_add(&odinmp_tacc_" + this + "_cleanup, odinmp_time() - odinmp_timer_cleanup_start);\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  }\n");
	    }
	    retval.append("  \n");
	}

        if (!nowait) {
	    
	    if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
		if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		    retval.append("  if (odinmp_timing) {\n");
		}
		retval.append("  odinmp_timer_barrier_start = odinmp_time();\n");
		if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		    retval.append("  }\n");
		}
		retval.append("  \n");
	    }
	    
            retval.append("  /* barrier at the end of parallel for \n");
            retval.append("     remove this section if nowait is specified */\n");
            retval.append("  odinmp_barrier(odinmp_team, \n");
            retval.append("                 &odinmp_team->for_barriers[" + index + "]);\n");
	    
	    if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
		if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		    retval.append("  if (odinmp_timing) {\n");
		}
		retval.append("  odinmp_tacc_add(&odinmp_tacc_" + this + "_barrier, odinmp_time() - odinmp_timer_barrier_start);\n");
		if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		    retval.append("  }\n");
		}
		retval.append("  \n");
	    }
	    
        }

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    retval.append("  \n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  if (odinmp_timing) {\n");
	    }
	    retval.append("  odinmp_tacc_add(&odinmp_tacc_" + this + "_total, odinmp_time() - odinmp_timer_init_start);\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  }\n");
	    }
	    retval.append("  \n");
	}

        retval.append("}\n");


        if (!dbv.lexicallyInParallel()) {
            retval.append("    } else {\n");

            retval.append("      for (" + attributesForSymbol(loopIndex).nameReplacement + " = " + lowerBound(dbv) + ";\n");
            retval.append("           " + attributesForSymbol(loopIndex).nameReplacement + " " + compOp + " " + bound(dbv) + ";\n");
            retval.append("           " + attributesForSymbol(loopIndex).nameReplacement + " += " + increment(dbv) + ")\n");

            dbv.visit(statement);
            retval.append(dbv.rv.toString());

            retval.append("    }\n");
            retval.append("  }\n");
        }

        return new String(retval);
    }


    public static String declareTiming() {
	StringBuffer retval = new StringBuffer();
	Enumeration e = inOrder.elements();
	while (e.hasMoreElements()) {
	    Object o = e.nextElement();
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_total;\n");
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_work;\n");
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_init;\n");
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_cleanup;\n");
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_init_iter;\n");
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_cleanup_iter;\n");
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_barrier;\n");
	}
	return new String(retval);
    }

    public static String externalDeclareTiming() {
	StringBuffer retval = new StringBuffer();
	Enumeration e = inOrder.elements();
	while (e.hasMoreElements()) {
	    Object o = e.nextElement();
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_total;\n");
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_work;\n");
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_init;\n");
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_cleanup;\n");
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_init_iter;\n");
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_cleanup_iter;\n");
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_barrier;\n");
	}
	return new String(retval);
    }

    public static String initTiming() {
	StringBuffer retval = new StringBuffer();
	Enumeration e = inOrder.elements();
	while (e.hasMoreElements()) {
	    Object o = e.nextElement();
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_total);\n");
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_work);\n");
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_init);\n");
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_cleanup);\n");
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_init_iter);\n");
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_cleanup_iter);\n");
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_barrier);\n");
	}
	return new String(retval);
    }

    public static String reportTiming() {
	StringBuffer retval = new StringBuffer();
	Enumeration e = inOrder.elements();
	while (e.hasMoreElements()) {
	    Object o = e.nextElement();
	    String s = o.toString();
	    
	    retval.append("   fprintf(stderr,\"For Construct " + s + ":\\n\");\n");
	    retval.append("   fprintf(stderr,\"  executed %d times total\\n\", odinmp_tacc_" + s + "_total.n);\n");
	    retval.append("   if (odinmp_tacc_" + s + "_total.n > 0) {\n");
	    retval.append("      fprintf(stderr,\"  total of %d iterations of work, %5f iterations/execution on average\\n\", odinmp_tacc_" + s + "_work.n, (double)odinmp_tacc_" + s + "_work.n / (double)odinmp_tacc_" + s + "_total.n);\n");
	    retval.append("      fprintf(stderr,\"  total time:         min = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_total.min);\n");
	    retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_total.total / odinmp_tacc_" + s + "_total.n);\n");
	    retval.append("      fprintf(stderr,\"                      max = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_total.max);\n");
	    retval.append("      fprintf(stderr,\"                    total = %#12.7g s accumulated\\n\", odinmp_tacc_" + s + "_total.total);\n");

	    retval.append("      fprintf(stderr,\"  working time:       min = %#12.7g s per iteration per execution\\n\", odinmp_tacc_" + s + "_work.min);\n");
	    retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per iteration per execution\\n\", odinmp_tacc_" + s + "_work.total / odinmp_tacc_" + s + "_work.n);\n");
	    retval.append("      fprintf(stderr,\"                      max = %#12.7g s per iteration per execution\\n\", odinmp_tacc_" + s + "_work.max);\n");
	    retval.append("      fprintf(stderr,\"                    total = %#12.7g s accumulated\\n\", odinmp_tacc_" + s + "_work.total);\n");

	    retval.append("      fprintf(stderr,\"  initialization:     min = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_init.min);\n");
	    retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_init.total / odinmp_tacc_" + s + "_init.n);\n");
	    retval.append("      fprintf(stderr,\"                      max = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_init.max);\n");
	    retval.append("      fprintf(stderr,\"                    total = %#12.7g s accumulated\\n\", odinmp_tacc_" + s + "_init.total);\n");

	    retval.append("      fprintf(stderr,\"  iter. init.ion:     min = %#12.7g s per iteration per execution\\n\", odinmp_tacc_" + s + "_init_iter.min);\n");
	    retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per iteration per execution\\n\", odinmp_tacc_" + s + "_init_iter.total / odinmp_tacc_" + s + "_init_iter.n);\n");
	    retval.append("      fprintf(stderr,\"                      max = %#12.7g s per iteration per execution\\n\", odinmp_tacc_" + s + "_init_iter.max);\n");
	    retval.append("      fprintf(stderr,\"                    total = %#12.7g s accumulated\\n\", odinmp_tacc_" + s + "_init_iter.total);\n");


	    retval.append("      fprintf(stderr,\"  cleanup:            min = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_cleanup.min);\n");
	    retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_cleanup.total / odinmp_tacc_" + s + "_cleanup.n);\n");
	    retval.append("      fprintf(stderr,\"                      max = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_cleanup.max);\n");
	    retval.append("      fprintf(stderr,\"                    total = %#12.7g s accumulated\\n\", odinmp_tacc_" + s + "_cleanup.total);\n");

	    retval.append("      fprintf(stderr,\"  iter. cleanup:      min = %#12.7g s per iteration per execution\\n\", odinmp_tacc_" + s + "_cleanup_iter.min);\n");
	    retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per iteration per execution\\n\", odinmp_tacc_" + s + "_cleanup_iter.total / odinmp_tacc_" + s + "_cleanup_iter.n);\n");
	    retval.append("      fprintf(stderr,\"                      max = %#12.7g s per iteration per execution\\n\", odinmp_tacc_" + s + "_cleanup_iter.max);\n");
	    retval.append("      fprintf(stderr,\"                    total = %#12.7g s accumulated\\n\", odinmp_tacc_" + s + "_cleanup_iter.total);\n");

	    if (!((ForConstructDescriptor)o).nowait) {
		retval.append("      fprintf(stderr,\"  barrier:            min = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_barrier.min);\n");
		retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_barrier.total / odinmp_tacc_" + s + "_cleanup.n);\n");
		retval.append("      fprintf(stderr,\"                      max = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_barrier.max);\n");
		retval.append("      fprintf(stderr,\"                    total = %#12.7g s accumulated\\n\", odinmp_tacc_" + s + "_barrier.total);\n");
	    } else {
		retval.append("      fprintf(stderr,\"  " + o + " has nowait specified, so no barrier at the end\\n\");\n");
	    }

	    retval.append("   }\n");
	    retval.append("   fprintf(stderr,\"\\n\");\n");
	}
	return new String(retval);
    }

}
