package se.lth.dit.odinmp;
import java.lang.*;
import java.util.*;

public class OmpSymbolAttributes {
    public static final int sharing_None = 0;
    public static final int sharing_Shared = 1;
    public static final int sharing_Private = 2;
    public static final int sharing_Firstprivate = 4;
    public static final int sharing_Lastprivate = 8;
    public static final int sharing_Reduction = 16;

    public int sharing = sharing_None;

    public String reductionOperator;

    public Hashtable attributes = new Hashtable();
    public String nameReplacement = null;
    public ConstructDescriptor constructDescriptor;

    public OmpSymbolAttributes(String nr, ConstructDescriptor cd) {
	sharing = sharing_None;

	nameReplacement = nr;
	constructDescriptor = cd;
    }

    public OmpSymbolAttributes(ConstructDescriptor cd) {
	this(null, cd);
    }

    public OmpSymbolAttributes() {
	this(null, null);
    }

    public void setSharing(int wantSharing) {
	switch(wantSharing) {
	case sharing_Shared:
	    if (sharing != sharing_None)
		return;
	    break;
	case sharing_Firstprivate:
	    wantSharing |= sharing_Private;
	    if ((sharing & sharing_Shared) != 0)
		return;
	    break;
	case sharing_Lastprivate:
	    wantSharing |= sharing_Private;
	    if ((sharing & sharing_Shared) != 0)
		return;
	    break;
	case sharing_Private:
	    if ((sharing & sharing_Shared) != 0)
		return;
	    break;
	case sharing_Reduction:
	    wantSharing |= sharing_Private;
	    if ((sharing & sharing_Shared) != 0)
		return;
	    break;
	}

	sharing |= wantSharing;
    }

    public String reductionInitValue() {
	if (reductionOperator == null)
	    return "((void *)0)";

	if (reductionOperator.equals("+")) {
	    return ("0");
	} 

	if (reductionOperator.equals("-")) {
	    return ("0");
	} 

	if (reductionOperator.equals("*")) {
	    return ("1");
	} 

	if (reductionOperator.equals("&")) {
	    return ("(~0)");
	} 

	if (reductionOperator.equals("|")) {
	    return ("0");
	} 

	if (reductionOperator.equals("^")) {
	    return ("0");
	} 

	if (reductionOperator.equals("&&")) {
	    return ("1");
	} 

	if (reductionOperator.equals("||")) {
	    return ("0");
	} 

	return "((void *)0)";
    }
}
