package se.lth.dit.odinmp;
import se.lth.dit.odinmp.*;
import se.lth.dit.odinmp.syntaxtree.*;
import se.lth.dit.odinmp.visitor.*;
import java.util.*;

public class ParallelClauseHandler extends SymbolClauseHandler {
    public ParallelConstructDescriptor parallelConstructDescriptor;


    public ParallelClauseHandler(SymbolFetcher sf, ParallelConstructDescriptor parCd) {
	super(sf);
	parallelConstructDescriptor = parCd;
    }

    public void handlePrivate(Vector v) {
	Vector syms = symbolsForNames(v);
	setSharingForSymbolsInConstruct(OmpSymbolAttributes.sharing_Private,
					syms,
					parallelConstructDescriptor);
	parallelConstructDescriptor.privateSymbols.addSymbols(syms);
    }

    public void handleFirstprivate(Vector v) {
	Vector syms = symbolsForNames(v);
	setSharingForSymbolsInConstruct(OmpSymbolAttributes.sharing_Firstprivate,
					syms,
					parallelConstructDescriptor);
	parallelConstructDescriptor.firstprivateSymbols.addSymbols(syms);
    }

    /*
      public void handleLastprivate(Vector v) {
      }
    */

    public void handleShared(Vector v) {
	Vector syms = symbolsForNames(v);
	setSharingForSymbolsInConstruct(OmpSymbolAttributes.sharing_Shared,
					syms,
					parallelConstructDescriptor);
	parallelConstructDescriptor.sharedSymbols.addSymbols(syms);
    }

    public void handleCopyin(Vector v) {
	Vector syms = symbolsForNames(v);
	parallelConstructDescriptor.copyinSymbols.addSymbols(syms);
    }

    public void handleDefault(String s) {
	if (s.equals("shared")) {
	    parallelConstructDescriptor.defaultSharing = OmpSymbolAttributes.sharing_Shared;
	} else if (s.equals("private")) {
	    parallelConstructDescriptor.defaultSharing = OmpSymbolAttributes.sharing_Private;
	} else if (s.equals("none")) {
	    parallelConstructDescriptor.defaultSharing = OmpSymbolAttributes.sharing_None;
	} else {
	    parallelConstructDescriptor.defaultSharing = OmpSymbolAttributes.sharing_None;
	}
    }

    public void handleReduction(String redop, Vector v) {
	Vector vv = symbolsForNames(v);
	Enumeration e = vv.elements();
	while (e.hasMoreElements()) {
	    Symbol s = (Symbol)e.nextElement();
	    OmpSymbolAttributes osa = parallelConstructDescriptor.attributesForSymbol(s);
	    osa.setSharing(OmpSymbolAttributes.sharing_Reduction);
	    osa.reductionOperator = redop;
	    parallelConstructDescriptor.reductionSymbols.addSymbol(s);

	    /* OdinMP.debugPrintln("reduction clause: osa = " + osa + ", symbol ``" + s + "'', red_op ``" + osa.reductionOperator + "''"); */
	    
	}
    }

    public void handleIf(Expression c) {
	parallelConstructDescriptor.condition = c;
    }

    /*
      public void handleNowait() {
      }
    */

    /*
      public void handleSchedule(String scheduleKind, String chunksize) {
      }
    */
}
