package se.lth.dit.odinmp;
import se.lth.dit.odinmp.*;
import se.lth.dit.odinmp.visitor.*;
import se.lth.dit.odinmp.syntaxtree.*;
import java.lang.*;
import java.util.*;

public class ParallelConstructDescriptor extends ConstructDescriptor {
    public static int nextIndex = 0;
    public static Vector inOrder = new Vector();

    public SymbolTable sharedSymbols = new SymbolTable();
    public SymbolTable privateSymbols = new SymbolTable();
    public SymbolTable copyinSymbols = new SymbolTable();
    public SymbolTable reductionSymbols = new SymbolTable();
    public SymbolTable firstprivateSymbols = new SymbolTable();
    public Expression condition = null;
    public int defaultSharing = OmpSymbolAttributes.sharing_None;
    public SourceFile sourceFile;

    public String extraCode;

    public SymbolTable threadDataSymbols = new SymbolTable();
    public SymbolTable sharedDataSymbols = new SymbolTable();

    public Statement statement = null;
    public ConstructDescriptor constructDescriptor = null;


    public ParallelConstructDescriptor() {
	super(1, "par", nextIndex++);
	inOrder.addElement(this);	
    }

    public ParallelConstructDescriptor(Statement cs) {
	this();
	statement = cs;
    }

    public ParallelConstructDescriptor(ConstructDescriptor cd) {
	this();
	constructDescriptor = cd;
    }

    public void prepareToGenerateCode(DataBuildingVisitor dbv) {
	String otd = OdinMP.THREAD_INFO("odinmp_td");
	Enumeration e;
	Symbol sym;
	OmpSymbolAttributes osa;

	sourceFile = dbv.odinmp.currentSourceFile;

	/* OdinMP.debugPrintln(this + ": generating code"); */

	e = sharedSymbols.symbolsInOrder.elements();
	while (e.hasMoreElements()) {
	    sym = (Symbol)e.nextElement();
	    if (!dbv.symbolIsGlobal(sym)) {
		attributesForSymbol(sym).nameReplacement = "(*odinmp_shared_data." + this + "." + sym.name + ")";
	    }
	}

	e = firstprivateSymbols.symbolsInOrder.elements();
	while (e.hasMoreElements()) {
	    sym = (Symbol)e.nextElement();
	    if (privateSymbols.containsSymbol(sym))
		privateSymbols.removeSymbol(sym);
	    attributesForSymbol(sym).nameReplacement = sym.name;
	    sharedDataSymbols.addSymbol(sym);
	}	
	
	e = reductionSymbols.symbolsInOrder.elements();
	while (e.hasMoreElements()) {
	    sym = (Symbol)e.nextElement();
	    osa = attributesForSymbol(sym);

	    /* OdinMP.debugPrintln("reduction: osa = " + osa + ", symbol ``" + sym + "'', red_op ``" + osa.reductionOperator + "''"); */

	    osa.nameReplacement = "odinmp_td->" + this + "." + sym.name;
	    /* OdinMP.debugPrintln("  setting nameReplacement to ``" + osa.nameReplacement + "''"); */
	    threadDataSymbols.addSymbol(sym);
	}
    }

    public String generateCode(DataBuildingVisitor dbv, String internalCode) {
	String otd = OdinMP.THREAD_INFO("odinmp_td");
	Enumeration e;
	Symbol sym;
	OmpSymbolAttributes osa;
	StringBuffer extraCodeBuffer = new StringBuffer();
	StringBuffer retvalBuffer = new StringBuffer();

	retvalBuffer.append("{\n");
	retvalBuffer.append("  /* we need a loop counter  */\n");
	retvalBuffer.append("  int odinmp_i, odinmp_id;\n");
	retvalBuffer.append("\n");
	retvalBuffer.append("  /* allocate a team of threads */\n");
	retvalBuffer.append("  odinmp_team_t *odinmp_team;\n");
	retvalBuffer.append("  odinmp_team_list_t *odinmp_tl;\n");
	retvalBuffer.append("  odinmp_thread_data_t *odinmp_td, *odinmp_master_td;\n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    retvalBuffer.append("  double odinmp_timer_init_start, odinmp_timer_init_thread_start;\n");
	    retvalBuffer.append("  double odinmp_timer_reap_start, odinmp_timer_reap_common_start;\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retvalBuffer.append("  if (odinmp_timing) {\n");
	    }
	    retvalBuffer.append("  odinmp_timer_init_start = odinmp_time();\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retvalBuffer.append("  }\n");
	    }
	}

	retvalBuffer.append("\n");
	retvalBuffer.append("  /* initialize the pointers to the shared variables */\n");

	e = sharedSymbols.symbolsInOrder.elements();
	while (e.hasMoreElements()) {
	    sym = (Symbol)e.nextElement();
	    if (!dbv.symbolIsGlobal(sym)) {
		retvalBuffer.append("   odinmp_shared_data." + this + "." + sym.name + " = & " + dbv.nameReplacementForSymbol(sym) + ";\n");
		sharedDataSymbols.addSymbol(sym);
	    }
	}

	retvalBuffer.append("\n");
	retvalBuffer.append("  /* initialize pointers to firstprivate variables, if any */\n");

	e = firstprivateSymbols.symbolsInOrder.elements();
	while (e.hasMoreElements()) {
	    sym = (Symbol)e.nextElement();
	    retvalBuffer.append(" odinmp_shared_data." + this + "." + sym.name + " = & " + dbv.nameReplacementForSymbol(sym) + ";");
	}	
	retvalBuffer.append("\n");
	
	if (condition != null) {
	    retvalBuffer.append("  /* check the condition */\n" +
				"  if (");

	    /* condition */
	    dbv.visit(condition);
	    retvalBuffer.append(dbv.rv.toString());
	    /* I wonder if this will work */

	    retvalBuffer.append(") {\n");
	    retvalBuffer.append("    /* allocate a team of threads */\n");
	    retvalBuffer.append("    /* ok, let's allocate as many as we can get */\n");
	    retvalBuffer.append("    odinmp_team = odinmp_allocate_team(0);\n");
	    retvalBuffer.append("  } else {\n");
	    retvalBuffer.append("    /* allocate a team of threads */\n");
	    retvalBuffer.append("    /* ok, allocate only one */\n");
	    retvalBuffer.append("    odinmp_team = odinmp_allocate_team(1);\n");
	    retvalBuffer.append("  }\n");
	} else {
	    retvalBuffer.append("  /* allocate a team of threads */\n");
	    retvalBuffer.append("  /* ok, let's allocate as many as we can get */\n");
	    retvalBuffer.append("  odinmp_team = odinmp_allocate_team(0);\n");
	}
	retvalBuffer.append("    odinmp_master_td = odinmp_thread_datas[odinmp_team->ids[0]];\n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retvalBuffer.append("  if (odinmp_timing) {\n");
	    }
	    retvalBuffer.append("  odinmp_tacc_add(&odinmp_tacc_" + this + "_init_common, (odinmp_timer_init_thread_start = odinmp_time()) - odinmp_timer_init_start);\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retvalBuffer.append("  }\n");
	    }
	    retvalBuffer.append("\n");
	}

	retvalBuffer.append("  /* for each thread, initialize the private data, and dispatch\n");
	retvalBuffer.append("     the thread  */\n");
	retvalBuffer.append("  for (odinmp_i = 0; odinmp_i < odinmp_team->n; odinmp_i++) {\n");
	retvalBuffer.append("    odinmp_id = odinmp_team->ids[odinmp_i];\n");
	retvalBuffer.append("    odinmp_td = odinmp_thread_datas[odinmp_id];\n");
	retvalBuffer.append("\n");
	retvalBuffer.append("    " + OdinMP.PRINTF("\"team: thread %d, id = %d\\n\", odinmp_i, odinmp_id") + ";\n");
	retvalBuffer.append("\n");
	retvalBuffer.append("    " + otd + "->run_region = " + index + ";\n");
	retvalBuffer.append("    if (" + otd + "->team) {\n");
	retvalBuffer.append("	odinmp_tl =\n");
	retvalBuffer.append("	  (odinmp_team_list_t *)malloc(sizeof(odinmp_team_list_t));\n");
	retvalBuffer.append("	odinmp_tl->next = " + otd + "->teams;\n");
	retvalBuffer.append("	" + otd + "->teams = odinmp_tl;\n");
	retvalBuffer.append("    }\n");
	retvalBuffer.append("    " + otd + "->team = odinmp_team;\n");
	
	retvalBuffer.append("\n");
	retvalBuffer.append("    if (odinmp_i != 0) {\n");
	retvalBuffer.append("      /* copy data from master thread into those threadprivate*/\n");
	retvalBuffer.append("	   /* variables that are listed in the copyin clause, if any */\n");
	
	e = copyinSymbols.symbolsInOrder.elements();
	while (e.hasMoreElements()) {
	    sym = (Symbol)e.nextElement();
	    retvalBuffer.append("       *(odinmp_td->threadprivate." + sym.name + ") = *(odinmp_master_td->threadprivate." + sym.name + ");\n");
	}	

	retvalBuffer.append("    }\n");
	
	retvalBuffer.append("\n    /* initialize reduction variables */\n");

	e = reductionSymbols.symbolsInOrder.elements();
	while (e.hasMoreElements()) {
	    sym = (Symbol)e.nextElement();
	    osa = attributesForSymbol(sym);
	    retvalBuffer.append("   odinmp_td->" + this + "." + sym.name + " = " + osa.reductionInitValue() + ";\n");
	    /* OdinMP.debugPrintln("initializing " + this + " reduction variable ``" + sym.name + "'' to " + osa.reductionInitValue()); */
	}	
	

	retvalBuffer.append("\n");
	retvalBuffer.append("    odinmp_dispatch(odinmp_td);\n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retvalBuffer.append("  if (odinmp_timing)\n");
	    }
	    retvalBuffer.append("  {\n");
	    retvalBuffer.append("    double tmp = odinmp_time();\n");
	    retvalBuffer.append("    odinmp_tacc_add(&odinmp_tacc_" + this + "_init_thread, tmp - odinmp_timer_init_thread_start);\n");
	    retvalBuffer.append("    odinmp_timer_init_thread_start = tmp;\n");
	    retvalBuffer.append("  }\n");
	}

	retvalBuffer.append("  }\n");
	retvalBuffer.append("\n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retvalBuffer.append("  if (odinmp_timing) {\n");
	    }
	    retvalBuffer.append("    odinmp_tacc_add(&odinmp_tacc_" + this + "_init_common, odinmp_time() - odinmp_timer_init_thread_start);\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retvalBuffer.append("  }\n");
	    }
	    retvalBuffer.append("\n");
	}

	retvalBuffer.append("  odinmp_thread(odinmp_thread_datas[odinmp_team->ids[0]]);\n");
	retvalBuffer.append("\n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retvalBuffer.append("  if (odinmp_timing) {\n");
	    }
	    retvalBuffer.append("  odinmp_timer_reap_start = odinmp_time();\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retvalBuffer.append("  }\n");
	    }
	}

	retvalBuffer.append("\n");
	retvalBuffer.append("  /* wait for all other threads to finish running this region */\n");
	retvalBuffer.append("  for (odinmp_i = 0; odinmp_i < odinmp_team->n; odinmp_i++) {\n");
	
	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    retvalBuffer.append("  double odinmp_timer_reap_thread_start;\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retvalBuffer.append("  if (odinmp_timing) {\n");
	    }
	    retvalBuffer.append("  odinmp_timer_reap_thread_start = odinmp_time();\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retvalBuffer.append("  }\n");
	    }
	}

	retvalBuffer.append("    odinmp_id = odinmp_team->ids[odinmp_i];\n");
	retvalBuffer.append("    odinmp_td = odinmp_thread_datas[odinmp_id];\n");
	retvalBuffer.append("    if (odinmp_i != 0)\n");
	retvalBuffer.append("	odinmp_wait_finished(odinmp_td);\n");
	retvalBuffer.append("\n");
	retvalBuffer.append("\n");
	retvalBuffer.append("    " + otd + "->run_region = -1;\n");
	retvalBuffer.append("    if (" + otd + "->teams) {\n");
	retvalBuffer.append("	odinmp_tl = " + otd + "->teams;\n");
	retvalBuffer.append("	" + otd + "->teams = odinmp_tl->next;\n");
	retvalBuffer.append("	" + otd + "->team = odinmp_tl->team;\n");
	retvalBuffer.append("	free(odinmp_tl);\n");
	retvalBuffer.append("    } else\n");
	retvalBuffer.append("	" + otd + "->team = ((void *)0);\n");
	retvalBuffer.append("\n");
	retvalBuffer.append("    /* reduce the reduction variables */\n");

	e = reductionSymbols.symbolsInOrder.elements();
	while (e.hasMoreElements()) {
	    sym = (Symbol)e.nextElement();
	    osa = attributesForSymbol(sym);
	    String nr = osa.nameReplacement;
	    osa.nameReplacement = null;

	    /* OdinMP.debugPrintln("reduction: osa = " + osa + ", symbol ``" + sym + "'', red_op ``" + osa.reductionOperator + "''"); */
	    
	    String nnr = dbv.nameReplacementForSymbol(sym);
	    retvalBuffer.append("   " + nnr + " = " + nnr + " " + osa.reductionOperator + " odinmp_td->" + this + "." + sym.name + ";\n");
	    osa.nameReplacement = nr;
	    /* OdinMP.debugPrintln("  resetting nameReplacement from ``" + nnr + "'' to ``" + osa.nameReplacement + "''"); */
	}	


	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retvalBuffer.append("  if (odinmp_timing) {\n");
	    }
	    retvalBuffer.append("  odinmp_tacc_add(&odinmp_tacc_" + this + "_reap_thread, odinmp_time() - odinmp_timer_reap_thread_start);\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retvalBuffer.append("  }\n");
	    }
	}

	retvalBuffer.append("  }\n");
	retvalBuffer.append("\n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retvalBuffer.append("  if (odinmp_timing) {\n");
	    }
	    retvalBuffer.append("    odinmp_timer_reap_common_start = odinmp_time();\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retvalBuffer.append("  }\n");
	    }
	}

	retvalBuffer.append("  /* deallocate team of threads */\n");
	retvalBuffer.append("  odinmp_free_team(odinmp_team);\n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retvalBuffer.append("  if (odinmp_timing)\n");
	    }
	    retvalBuffer.append("  {\n");
	    retvalBuffer.append("    double tmp = odinmp_time();");
	    retvalBuffer.append("    odinmp_tacc_add(&odinmp_tacc_" + this + "_reap_common, tmp - odinmp_timer_reap_common_start);\n");
	    retvalBuffer.append("    odinmp_tacc_add(&odinmp_tacc_" + this + "_total, tmp - odinmp_timer_init_start);\n");
	    retvalBuffer.append("  }\n");
	    retvalBuffer.append("\n");
	}

	retvalBuffer.append("}\n");


	/* Now, generate the extra code */
	/* OdinMP.debugPrintln(this + ": generating extra code"); */

	extraCodeBuffer.append("/* outer scope: variables */\n");
	extraCodeBuffer.append("{\n");
	extraCodeBuffer.append("  /* declare & define variables here */\n");

	e = firstprivateSymbols.symbolsInOrder.elements();
	while (e.hasMoreElements()) {
	    sym = (Symbol)e.nextElement();
	    if (privateSymbols.containsSymbol(sym))
		privateSymbols.removeSymbol(sym);
		extraCodeBuffer.append("  " + sym.nonStatic() + " = (*odinmp_shared_data." + this + "." + sym.name + ");\n");
	}

	e = privateSymbols.symbolsInOrder.elements();
	while (e.hasMoreElements()) {
	    sym = (Symbol)e.nextElement();
	    extraCodeBuffer.append("  " + sym.nonStatic() + ";\n");
	}

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    extraCodeBuffer.append("  double odinmp_timer_startTime;\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		extraCodeBuffer.append("  if (odinmp_timing) {\n");
	    }

	    extraCodeBuffer.append("  odinmp_timer_startTime = odinmp_time();\n");

	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		extraCodeBuffer.append("  }\n");
	    }
	}

	extraCodeBuffer.append(internalCode + "\n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		extraCodeBuffer.append("  if (odinmp_timing) {\n");
	    }

	    extraCodeBuffer.append("  odinmp_tacc_add(&odinmp_tacc_" + this + "_work, odinmp_time() - odinmp_timer_startTime);\n");

	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		extraCodeBuffer.append("  }\n");
	    }
	}

	extraCodeBuffer.append("}\n");
	extraCode = new String(extraCodeBuffer);

	/* OdinMP.debugPrintln(this + ": finished generating extra code"); */
	/* OdinMP.debugPrintln(this + ": finished generating code"); */

	cleanAffectedSymbols();

	return new String(retvalBuffer);
    }


    public String code(DataBuildingVisitor dbv) {
	String s = null;
	
	prepareToGenerateCode(dbv);

	dbv.constructDescriptorStack.push(this);

	if (constructDescriptor != null) {
	    s = constructDescriptor.code(dbv);
	} else if (statement != null) {
	    dbv.visit(statement);
	    s = dbv.rv.toString();
	} else {
	    // huh ? should never happen
	    s = "/* No generated code, should never happen */\n";
	}

	dbv.constructDescriptorStack.pop();

	return generateCode(dbv, s);
    }

    public String extraCode() {
	return extraCode;
    }

    public String threadDataTypedef() {
	if (threadDataSymbols.symbolsInOrder.size() > 0) {
	    Vector decSpecs = new Vector();
	    String s;
	    decSpecs.addElement("typedef");
	    decSpecs.addElement("struct");
	    
	    Symbol sym = new Symbol(decSpecs);
	    sym.add("{\n");
	    
	    Enumeration e = threadDataSymbols.symbolsInOrder.elements();
	    while (e.hasMoreElements()) {
		Symbol symbol = (Symbol)e.nextElement();
		sym.add(symbol.nonStatic().toString() + ";\n");
	    }
	    
	    sym.add("}");
	    sym.setName("odinmp_" + this + "_prm_t");
	    return sym.toString() + ";\n";
	} 
	return "";
    }

    public String threadData() {
	if (threadDataSymbols.symbolsInOrder.size() > 0) {
	    return "odinmp_" + this + "_prm_t " + this + ";\n";
	}
	return "";
    }

    public String sharedDataTypedef() {
	if (sharedDataSymbols.symbolsInOrder.size() > 0) {
	    Vector decSpecs = new Vector();
	    String s;
	    decSpecs.addElement("typedef");
	    decSpecs.addElement("struct");
	    
	    Symbol sym = new Symbol(decSpecs);
	    sym.add("{\n");
	    
	    Enumeration e = sharedDataSymbols.symbolsInOrder.elements();
	    while (e.hasMoreElements()) {
		Symbol symbol = (Symbol)e.nextElement();
		String ts = symbol.nonStatic().nonArrayTypeString("(*" + symbol.name + ")");
		sym.add(ts + ";\n");
	    }
	    
	    sym.add("}");
	    sym.setName("odinmp_" + this + "_shm_t");
	    return sym.toString() + ";\n";
	}
	return "";
    }

    public String sharedData() {
	if (sharedDataSymbols.symbolsInOrder.size() > 0) {
	    return "odinmp_" + this + "_shm_t " + this + ";\n";
	}
	return "";
    }

    public Vector dependsOnTypes() {
	Vector result = new Vector();
	Enumeration e;

	/* OdinMP.debugPrintln(this + ": generating type dependencies"); */

	e = sharedDataSymbols.symbolsInOrder.elements();
	while (e.hasMoreElements()) {
	    OdinMP.addToVector(result, ((Symbol)e.nextElement()).dependsOnTypes());
	}

	e = threadDataSymbols.symbolsInOrder.elements();
	while (e.hasMoreElements()) {
	    OdinMP.addToVector(result, ((Symbol)e.nextElement()).dependsOnTypes());
	}
	return result;
    }

    public static String declareTiming() {
	StringBuffer retval = new StringBuffer();
	Enumeration e = inOrder.elements();
	while (e.hasMoreElements()) {
	    Object o = e.nextElement();
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_total;\n");
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_work;\n");
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_init_common;\n");
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_init_thread;\n");
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_reap_common;\n");
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_reap_thread;\n");
	}
	return new String(retval);
    }

    public static String externalDeclareTiming() {
	StringBuffer retval = new StringBuffer();
	Enumeration e = inOrder.elements();
	while (e.hasMoreElements()) {
	    Object o = e.nextElement();
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_total;\n");
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_work;\n");
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_init_common;\n");
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_init_thread;\n");
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_reap_common;\n");
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_reap_thread;\n");
	}
	return new String(retval);
    }

    public static String initTiming() {
	StringBuffer retval = new StringBuffer();
	Enumeration e = inOrder.elements();
	while (e.hasMoreElements()) {
	    Object o = e.nextElement();
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_total);\n");
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_work);\n");
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_init_common);\n");
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_init_thread);\n");
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_reap_common);\n");
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_reap_thread);\n");
	}
	return new String(retval);
    }

    public static String reportTiming() {
	StringBuffer retval = new StringBuffer();
	Enumeration e = inOrder.elements();
	while (e.hasMoreElements()) {
	    Object o = e.nextElement();
	    String s = o.toString();
	    
	    retval.append("   fprintf(stderr,\"Parallel Construct " + s + ":\\n\");\n");
	    retval.append("   fprintf(stderr,\"  executed %d times total, with %f threads on average\\n\", odinmp_tacc_" + s + "_total.n, (double)odinmp_tacc_" + s + "_work.n / (double)odinmp_tacc_" + s + "_total.n);\n");
	    retval.append("   if (odinmp_tacc_" + s + "_total.n > 0) {\n");
	    retval.append("      fprintf(stderr,\"  total time:         min = %#12.7g s per iteration\\n\", odinmp_tacc_" + s + "_total.min);\n");
	    retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per iteration\\n\", odinmp_tacc_" + s + "_total.total / odinmp_tacc_" + s + "_total.n);\n");
	    retval.append("      fprintf(stderr,\"                      max = %#12.7g s per iteration\\n\", odinmp_tacc_" + s + "_total.max);\n");
	    retval.append("      fprintf(stderr,\"                    total = %#12.7g s accumulated\\n\", odinmp_tacc_" + s + "_total.total);\n");

	    retval.append("      fprintf(stderr,\"  working time:       min = %#12.7g s per thread per iteration\\n\", odinmp_tacc_" + s + "_work.min);\n");
	    retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per thread per iteration\\n\", odinmp_tacc_" + s + "_work.total / odinmp_tacc_" + s + "_work.n);\n");
	    retval.append("      fprintf(stderr,\"                      max = %#12.7g s per thread per iteration\\n\", odinmp_tacc_" + s + "_work.max);\n");
	    retval.append("      fprintf(stderr,\"                    total = %#12.7g s accumulated\\n\", odinmp_tacc_" + s + "_work.total);\n");

	    retval.append("      fprintf(stderr,\"  common init.ion:    min = %#12.7g s per iteration\\n\", odinmp_tacc_" + s + "_init_common.min);\n");
	    retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per iteration\\n\", odinmp_tacc_" + s + "_init_common.total / odinmp_tacc_" + s + "_init_common.n);\n");
	    retval.append("      fprintf(stderr,\"                      max = %#12.7g s per iteration\\n\", odinmp_tacc_" + s + "_init_common.max);\n");
	    retval.append("      fprintf(stderr,\"                    total = %#12.7g s accumulated\\n\", odinmp_tacc_" + s + "_init_common.total);\n");

	    retval.append("      fprintf(stderr,\"  thread init.ion:    min = %#12.7g s per thread per iteration\\n\", odinmp_tacc_" + s + "_init_thread.min);\n");
	    retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per thread per iteration\\n\", odinmp_tacc_" + s + "_init_thread.total / odinmp_tacc_" + s + "_init_thread.n);\n");
	    retval.append("      fprintf(stderr,\"                      max = %#12.7g s per thread per iteration\\n\", odinmp_tacc_" + s + "_init_thread.max);\n");
	    retval.append("      fprintf(stderr,\"                    total = %#12.7g s accumulated\\n\", odinmp_tacc_" + s + "_init_thread.total);\n");


	    retval.append("      fprintf(stderr,\"  common cleanup:     min = %#12.7g s per iteration\\n\", odinmp_tacc_" + s + "_reap_common.min);\n");
	    retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per iteration\\n\", odinmp_tacc_" + s + "_reap_common.total / odinmp_tacc_" + s + "_reap_common.n);\n");
	    retval.append("      fprintf(stderr,\"                      max = %#12.7g s per iteration\\n\", odinmp_tacc_" + s + "_reap_common.max);\n");
	    retval.append("      fprintf(stderr,\"                    total = %#12.7g s accumulated\\n\", odinmp_tacc_" + s + "_reap_common.total);\n");

	    retval.append("      fprintf(stderr,\"  thread cleanup:     min = %#12.7g s per thread per iteration\\n\", odinmp_tacc_" + s + "_reap_thread.min);\n");
	    retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per thread per iteration\\n\", odinmp_tacc_" + s + "_reap_thread.total / odinmp_tacc_" + s + "_reap_thread.n);\n");
	    retval.append("      fprintf(stderr,\"                      max = %#12.7g s per thread per iteration\\n\", odinmp_tacc_" + s + "_reap_thread.max);\n");
	    retval.append("      fprintf(stderr,\"                    total = %#12.7g s accumulated\\n\", odinmp_tacc_" + s + "_reap_thread.total);\n");

	    retval.append("      fprintf(stderr,\"  performed %#12.7g s of work in %#12.7g s of time\\n\", odinmp_tacc_" + s + "_work.total, odinmp_tacc_" + s + "_total.total);\n");
	    retval.append("      fprintf(stderr,\"  with on average %#12.7g threads,\\n  and a measured speedup of %#12.7g\\n\", (double)odinmp_tacc_" + s + "_work.n / (double)odinmp_tacc_" + s + "_total.n, odinmp_tacc_" + s + "_work.total / odinmp_tacc_" + s + "_total.total);\n");
	    //retval.append("      fprintf(stderr,\"  overhead:          time = %#12.7g s\\n\", 2 * (odinmp_tacc_" + s + "_total.total - odinmp_tacc_" + s + "_work.total));\n");
	    //retval.append("      fprintf(stderr,\"                    prop. = %#12.7g %%\\n\", 100.0 * ((odinmp_tacc_" + s + "_total.total / odinmp_tacc_" + s + "_work.total) - 1.0));\n");
	    retval.append("   }\n");
	    retval.append("   fprintf(stderr,\"\\n\");\n");
	}
	return new String(retval);
    }

}
