package se.lth.dit.odinmp;
import se.lth.dit.odinmp.*;
import se.lth.dit.odinmp.syntaxtree.*;
import se.lth.dit.odinmp.visitor.*;
import java.lang.*;
import java.util.*;

public class SectionsConstructDescriptor extends ConstructDescriptor {
    public static int nextIndex = 0;

    public static Vector inOrder = new Vector();

    public Vector sections = new Vector();
    public SymbolTable privateSymbols = new SymbolTable();
    public SymbolTable reductionSymbols = new SymbolTable();
    public SymbolTable firstprivateSymbols = new SymbolTable();
    public SymbolTable lastprivateSymbols = new SymbolTable();
    public SourceFile sourceFile;

    public boolean ordered = false;
    public boolean nowait = false;

    public SectionsConstructDescriptor() {
	super(3, "sect", nextIndex++);
	inOrder.addElement(this);
    }

    public String code(DataBuildingVisitor dbv) {
	Enumeration e;
	Symbol sym;
	SymbolTable declaredSymbols = new SymbolTable();
	OmpSymbolAttributes osa;
	int nSections = sections.size();
	int maxSection = nSections - 1;

	String otd = OdinMP.THREAD_INFO("odinmp_td");
	String owf = "odinmp_wsc_info.u.wsc_sect";
	String oll = "odinmp_wsc_info.u.wsc_sect.lap_list";
	sourceFile = dbv.odinmp.currentSourceFile;

	/* OdinMP.debugPrintln(this + ": generating code"); */

	StringBuffer retval = new StringBuffer();

	if (!dbv.lexicallyInParallel()) {
	    retval.append("  {\n");
	    retval.append("    odinmp_thread_data_t *odinmp_td = odinmp_self_data();\n");
	    retval.append("    if(((odinmp_thread_info_t *)odinmp_td)->team != ((void *)0)) {\n");
	    retval.append("\n");
	}
	
	retval.append("{\n");
	retval.append("  /* private variables; \n");
	retval.append("     if a variable is also firstprivate, then initialize it\n");
	retval.append("     from the corresponding outer variable\n");
	retval.append("     if a variable is a reduction variable, then initialize it\n");
	retval.append("     corresponding to the reduction operator */\n");

	/* Firstprivate variables */
	e = firstprivateSymbols.symbolsInOrder.elements();
	while (e.hasMoreElements()) {
	    sym = (Symbol)e.nextElement();
	    if (declaredSymbols.containsSymbol(sym))
		continue;
	    dbv.checkSharingForSymbol(sym);
	    String newName = "odinmp_flp_" + sym.name;
	    String nr = dbv.nameReplacementForSymbol(sym);
	    retval.append("  " + sym.nonStatic().typeString(newName) + " = " + nr + ";\n\n");
	    attributesForSymbol(sym).nameReplacement = newName;
	    declaredSymbols.addSymbol(sym);
	}
	
	/* Lastprivate variables */
	e = lastprivateSymbols.symbolsInOrder.elements();
	while (e.hasMoreElements()) {
	    sym = (Symbol)e.nextElement();
	    if (declaredSymbols.containsSymbol(sym))
		continue;
	    dbv.checkSharingForSymbol(sym);
	    String newName = "odinmp_flp_" + sym.name;
	    retval.append("  " + sym.nonStatic().typeString(newName) + ";\n");
	    attributesForSymbol(sym).nameReplacement = newName;
	    declaredSymbols.addSymbol(sym);
	}

	/* Reduction variables */
	e = reductionSymbols.symbolsInOrder.elements();
	while (e.hasMoreElements()) {
	    sym = (Symbol)e.nextElement();
	    if (declaredSymbols.containsSymbol(sym))
		continue;
	    dbv.checkSharingForSymbol(sym);
	    osa = attributesForSymbol(sym);
	    String newName = "odinmp_rdx_" + sym.name;
	    retval.append("  " + sym.nonStatic().typeString(newName) + " = " + osa.reductionInitValue() + ";\n");
	    osa.nameReplacement = newName;
	    declaredSymbols.addSymbol(sym);
	}

	/* "other" private variables */
	e = privateSymbols.symbolsInOrder.elements();
	while (e.hasMoreElements()) {
	    sym = (Symbol)e.nextElement();
	    if (declaredSymbols.containsSymbol(sym))
		continue;
	    osa = attributesForSymbol(sym);
	    String newName = sym.name;
	    retval.append("  " + sym + ";\n");
	    osa.nameReplacement = newName;
	    declaredSymbols.addSymbol(sym);
	}

	retval.append("  \n");
	retval.append("  /* odinmp variables */\n");
	retval.append("  /* wsc info etc */\n");
	retval.append("  odinmp_sect_slice_t odinmp_slice;\n");
	retval.append("  odinmp_wsc_info_t odinmp_wsc_info;\n");
        retval.append("  int odinmp_ttid = " + otd + "->team_thread_id.id;\n");
	retval.append("\n");
	retval.append("  /* for slice fetching */\n");
	retval.append("  odinmp_team_t *odinmp_team = ((odinmp_thread_info_t *) (odinmp_td))->team;\n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    retval.append("  double odinmp_timer_init_start, odinmp_timer_init_iter_start, odinmp_timer_work_start, odinmp_timer_cleanup_iter_start, odinmp_timer_cleanup_start, odinmp_timer_barrier_start;\n");

	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  if (odinmp_timing) {\n");
	    }
	    retval.append("  odinmp_timer_init_start = odinmp_time();\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  }\n");
	    }
	    retval.append("  \n");
	}

	retval.append("  /* record info about this work sharing construct */\n");
	retval.append("  odinmp_wsc_info.next = " + otd + "->wsc_info;\n");
	retval.append("  " + otd + "->wsc_info = &odinmp_wsc_info;\n");
	retval.append("  odinmp_wsc_info.type = odinmp_wsc_type_sect;\n");

	retval.append("  \n");
	retval.append("  ODINMP_TIMING_WAIT_BEGIN\n");
	retval.append("  pthread_mutex_lock(&odinmp_team->lock);\n");
	retval.append("  ODINMP_TIMING_WAIT_END\n");
	retval.append("\n");
	retval.append("  for (" + oll + " = odinmp_team->sect_laps[" + index + "];\n");
	retval.append("       " + oll + " && " + oll + "->lap > odinmp_team->thread_lap_infos[odinmp_ttid]." + this + ".lap;\n");
	retval.append("       " + oll + " = " + oll + "->next)\n");
	retval.append("    ;\n");
	retval.append("\n");
	retval.append("  if (!" + oll + " || " + oll + "->lap != odinmp_team->thread_lap_infos[odinmp_ttid]." + this + ".lap) {\n");
	retval.append("    " + oll + " = (odinmp_sect_lap_list_t *)\n");
	retval.append("      malloc(sizeof(odinmp_sect_lap_list_t));\n");
	retval.append("    " + oll + "->lap = odinmp_team->thread_lap_infos[odinmp_ttid]." + this + ".lap;\n");
	retval.append("    " + oll + "->n_left_to_leave = odinmp_team->n;\n");
	retval.append("    " + oll + "->sect_data.n_left = " + nSections + ";\n");
	retval.append("    " + oll + "->next = odinmp_team->sect_laps[" + index + "];\n");
	retval.append("    odinmp_team->sect_laps[" + index + "] = " + oll + ";\n");
	retval.append("  }\n");
	retval.append("\n");
	retval.append("    pthread_mutex_unlock(&odinmp_team->lock);\n");
	retval.append("\n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  if (odinmp_timing) {\n");
	    }
	    retval.append("  odinmp_tacc_add(&odinmp_tacc_" + this + "_init, odinmp_time() - odinmp_timer_init_start);\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  }\n");
	    }
	    retval.append("  \n");
	}

	retval.append("  while (1) {\n");
	retval.append("\n");
	retval.append("    " + OdinMP.PRINTF("\"thread %d to fetch a slice of sections " + index + "\\n\", " + otd + "->id") + ";\n");
	retval.append("          \n");
	retval.append("\n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  if (odinmp_timing) {\n");
	    }
	    retval.append("    odinmp_timer_init_iter_start = odinmp_time();\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  }\n");
	    }
	    retval.append("  \n");
	}

	retval.append("    /* reset slice */\n");
	retval.append("    odinmp_slice.last = 0;\n");
	retval.append("    odinmp_slice.done = 0;\n");
	retval.append("\n");
	retval.append("    ODINMP_TIMING_WAIT_BEGIN\n");
	retval.append("    pthread_mutex_lock(&odinmp_team->lock);\n");
	retval.append("    ODINMP_TIMING_WAIT_END\n");
	retval.append("\n");
	retval.append("    if (" + oll + "->sect_data.n_left > 0) {\n");
	retval.append("      odinmp_slice.section = " + nSections + " - (" + oll + "->sect_data.n_left--);\n");
	retval.append("      odinmp_slice.done = 0;\n");
	retval.append("      /* are we performing the sequentially last section? */\n");
	retval.append("      odinmp_slice.last = (odinmp_slice.section == " + maxSection + ");\n");
	retval.append("    } else {\n");
	retval.append("      odinmp_slice.section = 0;\n");
	retval.append("      odinmp_slice.done = 1;\n");
	retval.append("      odinmp_slice.last = 0;\n");
	retval.append("    }\n");
	retval.append("\n");
	retval.append("    pthread_mutex_unlock(&odinmp_team->lock);\n");
	retval.append("\n");
	retval.append("    " + OdinMP.PRINTF("\"thread %d got slice %d of sections " + index + "\\n\", " + otd + "->id, odinmp_slice.section") + ";\n");
	retval.append("            \n");
	retval.append("    if (odinmp_slice.done) {\n");
	retval.append("      /* increment the lap counter */\n");
	retval.append("      odinmp_team->thread_lap_infos[odinmp_ttid]." + this + ".lap ++;\n");
	retval.append("      /* ... and exit the while(1) loop */\n");
	retval.append("      break;\n");
	retval.append("    }\n");
	retval.append("\n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  if (odinmp_timing) {\n");
	    }
	    retval.append("  odinmp_tacc_add(&odinmp_tacc_" + this + "_init_iter, (odinmp_timer_work_start = odinmp_time()) - odinmp_timer_init_iter_start);\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  }\n");
	    }
	    retval.append("\n");
	}

	retval.append("    switch (odinmp_slice.section) {\n");

	dbv.constructDescriptorStack.push(this);	
	for (int i = 0; i < nSections; i++) {
	    retval.append("    case " + i + ":\n");
	    Statement cs = (Statement)sections.elementAt(i);
	    dbv.visit(cs);
	    retval.append(dbv.rv.toString() + "\n");
	    retval.append("      break;\n");
	}
	dbv.constructDescriptorStack.pop();

	retval.append("    default:\n");
	retval.append("      " + OdinMP.PRINTF("\" " + this + ": don't have a \\\"section\\\" %d\\n\", odinmp_slice.section") + ";\n");
	retval.append("    }\n");
	retval.append("\n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  if (odinmp_timing) {\n");
	    }
	    retval.append("  odinmp_tacc_add(&odinmp_tacc_" + this + "_work, (odinmp_timer_cleanup_iter_start = odinmp_time()) - odinmp_timer_work_start);\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  }\n");
	    }
	    retval.append("  \n");
	}

	retval.append("\n");
	retval.append("    /* did we just execute the last section? */\n");
	retval.append("    if (odinmp_slice.last) {\n");
	retval.append("      /* yes we did, copy out the lastprivate variables */\n");
	/* copy data out to lastprivate variables */
	e = lastprivateSymbols.symbolsInOrder.elements();
	while (e.hasMoreElements()) {
	    sym = (Symbol)e.nextElement();
	    osa = attributesForSymbol(sym);
	    String newName = osa.nameReplacement;
	    osa.nameReplacement = null;
	    retval.append("    " + dbv.nameReplacementForSymbol(sym) + " = " + newName + ";\n");
	}

	retval.append("    }\n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  if (odinmp_timing) {\n");
	    }
	    retval.append("  odinmp_tacc_add(&odinmp_tacc_" + this + "_cleanup_iter, odinmp_time() - odinmp_timer_cleanup_iter_start);\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  }\n");
	    }
	    retval.append("  \n");
	}

	retval.append("  } /* while(1) */\n");
	retval.append("\n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  if (odinmp_timing) {\n");
	    }
	    retval.append("  odinmp_tacc_add(&odinmp_tacc_" + this + "_init_iter, (odinmp_timer_cleanup_start = odinmp_time()) - odinmp_timer_init_iter_start);\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  }\n");
	    }
	    retval.append("  \n");
	}

	retval.append("  /* restore work sharing construct info */\n");
	retval.append("  " + otd + "->wsc_info = odinmp_wsc_info.next;\n");
	retval.append("\n");

	retval.append("    ODINMP_TIMING_WAIT_BEGIN\n");
	retval.append("    pthread_mutex_lock(&odinmp_team->lock);\n");
	retval.append("    ODINMP_TIMING_WAIT_END\n");
	retval.append("\n");
	retval.append("    /* reduce reduction variables */\n");
	/* reduce reduction variables */
	e = reductionSymbols.symbolsInOrder.elements();
	if (e.hasMoreElements()) {
	    while (e.hasMoreElements()) {
		sym = (Symbol)e.nextElement();
		osa = attributesForSymbol(sym);
		String newName = osa.nameReplacement;
		osa.nameReplacement = null;
		String nr = dbv.nameReplacementForSymbol(sym);
		retval.append("    " + nr + " = " + nr + " " + osa.reductionOperator + " " + newName + ";\n");
	    }
	}

	retval.append("  /* update the lap list */\n");
	retval.append("  if (--" + oll + "->n_left_to_leave == 0) {\n");
	retval.append("    odinmp_sect_lap_list_t *oll = odinmp_team->sect_laps[" + index + "];\n");
	retval.append("    if (oll == " + oll + ") {\n");
	retval.append("      odinmp_team->sect_laps[" + index + "] = oll->next;\n");
	retval.append("    } else {\n");
	retval.append("      while (oll->next != " + oll + ")\n");
	retval.append("        oll = oll->next;\n");
	retval.append("      oll->next = " + oll + "->next;\n");
	retval.append("    }\n");
	retval.append("    free(oll);\n");
	retval.append("  }\n");
	retval.append("  pthread_mutex_unlock(&odinmp_team->lock);\n");

	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  if (odinmp_timing) {\n");
	    }
	    retval.append("  odinmp_tacc_add(&odinmp_tacc_" + this + "_cleanup, odinmp_time() - odinmp_timer_cleanup_start);\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  }\n");
	    }
	    retval.append("  \n");
	}

	if (!nowait) {
	    
	    if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
		if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		    retval.append("  if (odinmp_timing) {\n");
		}
		retval.append("  odinmp_timer_barrier_start = odinmp_time();\n");
		if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		    retval.append("  }\n");
		}
		retval.append("  \n");
	    }
	    
	    retval.append("  /* barrier at the end of sections \n");
	    retval.append("     remove this if nowait is specified */\n");
	    retval.append("  odinmp_barrier(odinmp_team, \n");
	    retval.append("                 &odinmp_team->sect_barriers[" + index + "]);\n");
	    
	    if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
		if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		    retval.append("  if (odinmp_timing) {\n");
		}
		retval.append("  odinmp_tacc_add(&odinmp_tacc_" + this + "_barrier, odinmp_time() - odinmp_timer_barrier_start);\n");
		if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		    retval.append("  }\n");
		}
		retval.append("  \n");
	    }
	    
	}


	if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  if (odinmp_timing) {\n");
	    }
	    retval.append("  odinmp_tacc_add(&odinmp_tacc_" + this + "_total, odinmp_time() - odinmp_timer_init_start);\n");
	    if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		retval.append("  }\n");
	    }
	    retval.append("  \n");
	}

	retval.append("}\n");

	if (!dbv.lexicallyInParallel()) {
	    retval.append("    } else {\n");

	    for (int i = 0; i < nSections; i++) {
		Statement cs = (Statement)sections.elementAt(i);
		dbv.visit(cs);
		retval.append(dbv.rv.toString() + "\n");
	    }

	    retval.append("    }\n");
	    retval.append("  }\n");
	}

	String retString = new String(retval);

	/* OdinMP.debugPrintln(this + ": generated code = ``\n" + retString + "''"); */

	return retString;
    }



    public static String declareTiming() {
	StringBuffer retval = new StringBuffer();
	Enumeration e = inOrder.elements();
	while (e.hasMoreElements()) {
	    Object o = e.nextElement();
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_total;\n");
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_work;\n");
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_init;\n");
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_cleanup;\n");
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_init_iter;\n");
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_cleanup_iter;\n");
	    retval.append("odinmp_tacc_t odinmp_tacc_" + o + "_barrier;\n");
	}
	return new String(retval);
    }

    public static String externalDeclareTiming() {
	StringBuffer retval = new StringBuffer();
	Enumeration e = inOrder.elements();
	while (e.hasMoreElements()) {
	    Object o = e.nextElement();
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_total;\n");
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_work;\n");
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_init;\n");
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_cleanup;\n");
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_init_iter;\n");
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_cleanup_iter;\n");
	    retval.append("extern odinmp_tacc_t odinmp_tacc_" + o + "_barrier;\n");
	}
	return new String(retval);
    }

    public static String initTiming() {
	StringBuffer retval = new StringBuffer();
	Enumeration e = inOrder.elements();
	while (e.hasMoreElements()) {
	    Object o = e.nextElement();
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_total);\n");
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_work);\n");
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_init);\n");
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_cleanup);\n");
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_init_iter);\n");
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_cleanup_iter);\n");
	    retval.append("   odinmp_tacc_init(&odinmp_tacc_" + o + "_barrier);\n");
	}
	return new String(retval);
    }

    public static String reportTiming() {
	StringBuffer retval = new StringBuffer();
	Enumeration e = inOrder.elements();
	while (e.hasMoreElements()) {
	    Object o = e.nextElement();
	    String s = o.toString();
	    
	    retval.append("   fprintf(stderr,\"Sections Construct " + s + ":\\n\");\n");
	    retval.append("   fprintf(stderr,\"  executed %d times total\\n\", odinmp_tacc_" + s + "_total.n);\n");
	    retval.append("   if (odinmp_tacc_" + s + "_total.n > 0) {\n");
	    retval.append("      fprintf(stderr,\"  total of %d iterations of work, %5f iterations/execution on average\\n\", odinmp_tacc_" + s + "_work.n, (double)odinmp_tacc_" + s + "_work.n / (double)odinmp_tacc_" + s + "_total.n);\n");
	    retval.append("      fprintf(stderr,\"  total time:         min = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_total.min);\n");
	    retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_total.total / odinmp_tacc_" + s + "_total.n);\n");
	    retval.append("      fprintf(stderr,\"                      max = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_total.max);\n");
	    retval.append("      fprintf(stderr,\"                    total = %#12.7g s accumulated\\n\", odinmp_tacc_" + s + "_total.total);\n");

	    retval.append("      fprintf(stderr,\"  working time:       min = %#12.7g s per iteration per execution\\n\", odinmp_tacc_" + s + "_work.min);\n");
	    retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per iteration per execution\\n\", odinmp_tacc_" + s + "_work.total / odinmp_tacc_" + s + "_work.n);\n");
	    retval.append("      fprintf(stderr,\"                      max = %#12.7g s per iteration per execution\\n\", odinmp_tacc_" + s + "_work.max);\n");
	    retval.append("      fprintf(stderr,\"                    total = %#12.7g s accumulated\\n\", odinmp_tacc_" + s + "_work.total);\n");

	    retval.append("      fprintf(stderr,\"  initialization:     min = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_init.min);\n");
	    retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_init.total / odinmp_tacc_" + s + "_init.n);\n");
	    retval.append("      fprintf(stderr,\"                      max = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_init.max);\n");
	    retval.append("      fprintf(stderr,\"                    total = %#12.7g s accumulated\\n\", odinmp_tacc_" + s + "_init.total);\n");

	    retval.append("      fprintf(stderr,\"  iter. init.ion:     min = %#12.7g s per iteration per execution\\n\", odinmp_tacc_" + s + "_init_iter.min);\n");
	    retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per iteration per execution\\n\", odinmp_tacc_" + s + "_init_iter.total / odinmp_tacc_" + s + "_init_iter.n);\n");
	    retval.append("      fprintf(stderr,\"                      max = %#12.7g s per iteration per execution\\n\", odinmp_tacc_" + s + "_init_iter.max);\n");
	    retval.append("      fprintf(stderr,\"                    total = %#12.7g s accumulated\\n\", odinmp_tacc_" + s + "_init_iter.total);\n");


	    retval.append("      fprintf(stderr,\"  cleanup:            min = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_cleanup.min);\n");
	    retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_cleanup.total / odinmp_tacc_" + s + "_cleanup.n);\n");
	    retval.append("      fprintf(stderr,\"                      max = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_cleanup.max);\n");
	    retval.append("      fprintf(stderr,\"                    total = %#12.7g s accumulated\\n\", odinmp_tacc_" + s + "_cleanup.total);\n");

	    retval.append("      fprintf(stderr,\"  iter. cleanup:      min = %#12.7g s per iteration per execution\\n\", odinmp_tacc_" + s + "_cleanup_iter.min);\n");
	    retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per iteration per execution\\n\", odinmp_tacc_" + s + "_cleanup_iter.total / odinmp_tacc_" + s + "_cleanup_iter.n);\n");
	    retval.append("      fprintf(stderr,\"                      max = %#12.7g s per iteration per execution\\n\", odinmp_tacc_" + s + "_cleanup_iter.max);\n");
	    retval.append("      fprintf(stderr,\"                    total = %#12.7g s accumulated\\n\", odinmp_tacc_" + s + "_cleanup_iter.total);\n");

	    if (!((SectionsConstructDescriptor)o).nowait) {
		retval.append("      fprintf(stderr,\"  barrier:            min = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_barrier.min);\n");
		retval.append("      fprintf(stderr,\"                      avg = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_barrier.total / odinmp_tacc_" + s + "_cleanup.n);\n");
		retval.append("      fprintf(stderr,\"                      max = %#12.7g s per execution\\n\", odinmp_tacc_" + s + "_barrier.max);\n");
		retval.append("      fprintf(stderr,\"                    total = %#12.7g s accumulated\\n\", odinmp_tacc_" + s + "_barrier.total);\n");
	    } else {
		retval.append("      fprintf(stderr,\"  " + o + " has nowait specified, so no barrier at the end\\n\");\n");
	    }

	    retval.append("   }\n");
	    retval.append("   fprintf(stderr,\"\\n\");\n");
	}
	return new String(retval);
    }
}
