package se.lth.dit.odinmp;

import java.util.*;
import java.io.*;
import se.lth.dit.odinmp.*;
import se.lth.dit.odinmp.syntaxtree.*;
import se.lth.dit.odinmp.visitor.*;

public class SourceFile {
    public String name;
    public String mangledName;
    public File file;
    public TranslationUnit root;
    public DataBuildingVisitor dbv;
    public String includes;
    public OdinMP odinmp;

    public SourceFile (OdinMP o, File f, String n) {
	odinmp = o;
	file = f;
	name = n;
	mangledName = SourceFile.mangle(name);
    }
    
    static boolean haveToMangle(char c) {
	boolean result = !
	    (
	     ('a' <= c && c <= 'z') ||
	     ('A' <= c && c <= 'Z') ||
	     ('0' <= c && c <= '9')
	     );

	/* OdinMP.debugPrint("char `" + c + "' "); */
	if(result) {
	    /* OdinMP.debugPrintln("needs to be mangled"); */
	} else {
	    /* OdinMP.debugPrintln("can be kept as it is"); */
	}
	
	return result;
    }

    static String mangle(String s) {
	StringBuffer sb = new StringBuffer(s);
	for (int i = 0; i < sb.length(); i++) {
	    char c = sb.charAt(i);
	    if (haveToMangle(c)) {
		sb.setCharAt(i, '_');
	    }
	}
	return new String(sb);
    }

    public TranslationUnit parse () {
	if (root == null) {
	    CParser parser = new CParser(preprocessed());
	    
	    // Hack to include type "special types"
	    parser.types.put("__signed__", Boolean.TRUE);
	    parser.types.put("__const", Boolean.TRUE);
	    parser.types.put("__inline__", Boolean.TRUE);
	    parser.types.put("__extension__", Boolean.TRUE);
	    parser.types.put("__inline", Boolean.TRUE);
	    parser.types.put("__signed", Boolean.TRUE);

	    // and parse it
	    try {
		root = parser.TranslationUnit();
	    }
	    catch (ParseException e) {
		System.err.println(OdinMP.ident + ":  Encountered errors during parse in file ``" + name + "''");
		e.printStackTrace();
		System.err.println(e.getMessage());
		return null;
	    }
	}
	return root;
    }

    public Reader preprocessed() {
	StringWriter buffer;
	final String includes;
	final String sourceData;
	final String preprocessedData;

	if (OdinMP.includes == null) {
	    buffer = new StringWriter();
	    try {
		char buf[] = new char[10240];
		FileReader fr = new FileReader(new File(OdinMP.supporting(), "odinmp_includes.h"));
		int nread;
		while ((nread = fr.read(buf)) > 0) {
		    buffer.write(buf, 0, nread);
		}
		fr.close();
		buffer.close();
	    } catch (Exception ex) {
		System.err.println("Caught exception trying to gather odinmp includes: " + ex);
		return null;
	    }
	    odinmp.includes = buffer.toString();
	}

	includes = new String(OdinMP.includes);

	buffer = new StringWriter();
	try {
	    char buf[] = new char[10240];
	    FileReader fr = new FileReader(file);
	    int nread;
	    while ((nread = fr.read(buf)) > 0) {
		buffer.write(buf, 0, nread);
	    }
	    fr.close();
	    buffer.close();
	} catch (Exception ex) {
	    System.err.println("Caught exception trying to preprocess file ``" + name + "'': " + ex);
	    return null;
	}
	sourceData = buffer.toString();
	buffer = null;

	try {
	    String cpp = OdinMP.cpp();
	    
	    OdinMP.debugPrintln("cpp = ``" + cpp + "''");
	    OdinMP.debugPrintln("cppOptions = ``" + OdinMP.cppOptions + "''");
	    
	    if (cpp.length() == 0)
		return new FileReader(file);

	    Process p = Runtime.getRuntime().exec(cpp + " -I " + OdinMP.supporting() + " " + OdinMP.cppOptions);
	    final OutputStreamWriter toCpp = new OutputStreamWriter(p.getOutputStream());
	    final InputStreamReader fromCpp = new InputStreamReader(p.getInputStream());
	    final InputStreamReader cppErrors = new InputStreamReader(p.getErrorStream());

	    Runnable feeder = new Runnable() {
		public void run() {	
		    try {
			toCpp.write(includes);
			char buf[] = new char[10240];
			FileReader fr = new FileReader(file);
			int nread;
			while ((nread = fr.read(buf)) > 0) {
			    toCpp.write(buf, 0, nread);
			}
			fr.close();
			toCpp.close();
		    } catch (Exception ex) {	
			System.err.println("Caught exception feeding cpp, trying to precompile file ``" + name + "'': " + ex);
			return;
		    }
		}
	    };
	    new Thread(feeder).start();

	    Runnable errorReader = new Runnable() {
		public void run() {	
		    StringWriter b = new StringWriter();
		    try {	
			int nread;
			char buf[] = new char[1024];
			while ((nread = cppErrors.read(buf)) > 0) {
			    b.write(buf, 0, nread);
			}
			cppErrors.close();
			b.close();
		    } catch (Exception ex) {	
			System.err.println("Caught exception feeding cpp, trying to precompile file ``" + name + "'': " + ex);
		    }
		    if (b.getBuffer().length() > 0) {
			System.err.println("--- cpp errors ---");
			System.err.println(b.toString());
			System.err.println("--- end of cpp errors ---");
		    }
		}
	    };
	    new Thread(errorReader).start();

	    buffer = new StringWriter();
	    int nread;
	    char buf[] = new char[10240];
	    while ((nread = fromCpp.read(buf)) > 0) {
		//System.err.println("muncher: read " + nread + " chars preprocessed ``" + name + "''");
		buffer.write(buf, 0, nread);
	    }
	    fromCpp.close();
	    buffer.close();
	    preprocessedData = buffer.toString();

	    //System.err.println("Preprocessed data = ``");
	    //System.err.println(preprocessedData);
	    //System.err.println("''");
	    //Thread.currentThread().sleep(2000);
	    //System.exit(0);

	    return new StringReader(preprocessedData);
	} catch (Exception ex) {
	    System.err.println("Caught exception trying to precompile file ``" + name + "'': " + ex);
	    return null;
	}
    }
}

