package se.lth.dit.odinmp;
import se.lth.dit.odinmp.*;
import se.lth.dit.odinmp.syntaxtree.*;
import java.lang.*;
import java.util.*;

public class Symbol {
    public Vector declarationSpecifiers;
    public String preName = "";
    public String name = "";
    public String postName = "";
    public Initializer initializer;
    public boolean inPreName = true;
    public SymbolTable symbolTable;
    public SymbolTable subSymbols = new SymbolTable();
    public Vector dependsOnTypes = new Vector();
    public boolean isThreadprivate = false;

    public Symbol (Vector decSpecs) {
	declarationSpecifiers = decSpecs;
    }

    public Symbol () {
	this(new Vector());
    }

    public Symbol nonStatic() {
	Vector newDecSpecs = (Vector)declarationSpecifiers.clone();
	newDecSpecs.removeElement("static");
	newDecSpecs.removeElement("extern");
	Symbol newSym = new Symbol(newDecSpecs);
	newSym.name = name;
	newSym.preName = preName;
	newSym.postName = postName;
	newSym.initializer = initializer;
	newSym.symbolTable = symbolTable;
	newSym.dependsOnTypes = dependsOnTypes;
	newSym.isThreadprivate = isThreadprivate;

	return newSym;
    }

    public void setName(String n) {
	/* OdinMP.debugPrintln("Symbol " + this + ": setting name to ``" + n + "''"); */
	name = n;
	inPreName = false;
    }

    public void add(String s) {
	if (inPreName) {
	    /* OdinMP.debugPrintln("Symbol " + this + ": adding ``" + s + "'' PRE name"); */
	    preName = preName + " " + s;
	} else {
	    /* OdinMP.debugPrintln("Symbol " + this + ": adding ``" + s + "'' POST name"); */
	    postName = postName + " " + s;
	}
    }

    public String declarationSpecifiersAsString() {
	String s = "";
	for (Enumeration e = declarationSpecifiers.elements();
	     e.hasMoreElements();
	     ) {
	    s = s + e.nextElement().toString() + " ";
	}
	return s;
    }

    public String toString() {
	return declarationSpecifiersAsString() + preName + " " + name + " " + postName;
    }

    public String typeString() {
	return declarationSpecifiersAsString() + preName + " " + postName;
    }

    public String typeString(String insteadOfName) {
	return declarationSpecifiersAsString() + preName + " " + insteadOfName + " " + postName;
    }

    public String nonArrayTypeString(String insteadOfName) {
	if (postName.endsWith("]")) { // it's an array -replace with 'Pointer To' type
	    String nPre = " * " + preName;
	    int ri = postName.lastIndexOf("[");
	    String nPost = postName.substring(0, ri);
	    return declarationSpecifiersAsString() + nPre + " " + insteadOfName + " " + nPost;
	} else {
	    return typeString(insteadOfName);
	}
    }

    public String nonArrayTypeString() {
	return nonArrayTypeString("");
    }

    public void dependOnType(Symbol type) {
	dependsOnTypes.addElement(type);
    }

    public void dependOnTypes(Vector v) {
	Enumeration e = v.elements();
	while (e.hasMoreElements()) {
	    Symbol type = (Symbol)e.nextElement();
	    dependOnType(type);
	}
    }

    public Vector dependsOnTypes() {
	Enumeration e1, e2;
	Vector result = new Vector();
	Vector v;
	e1 = dependsOnTypes.elements();
	while (e1.hasMoreElements()) {
	    Symbol type = (Symbol)e1.nextElement();

	    if (!result.contains(type)) {
		OdinMP.addToVector(result, type.dependsOnTypes());
		result.addElement(type);
	    }

	}
	return result;
    }
}
