package se.lth.dit.odinmp.visitor;
import se.lth.dit.odinmp.syntaxtree.*;
import se.lth.dit.odinmp.*;
import java.util.*;
import java.io.*;

public class DataBuildingVisitor extends ReturningVisitor implements SymbolFetcher {
    
    public SymbolTable typedefs = new SymbolTable();
    public SymbolTable threadprivateSymbols = new SymbolTable();
    public Stack savedDeclStack = null;
    public Stack declStack = new Stack();
    public Stack typedefSavedDeclStack = new Stack();
    public Stack symbolTableStack = new Stack();
    public int defaultSharing = OmpSymbolAttributes.sharing_Shared;
    public Stack constructDescriptorStack = new Stack();
    public ParallelConstructDescriptor topmostParallelConstruct = null;
    public ConstructDescriptor topmostWorkSharingConstruct = null;
    public SymbolTable functionDefinitionSymbolTable = null;
    public ClauseHandler clauseHandler = null;
    public OdinMP odinmp = null;
    public String sep = " ";
    public Vector parallelRegions = new Vector();
    public boolean mainInThisFile = false;
    public boolean workingOnSymbol = false;
    public SourceFile sourceFile = null;

    Vector variableList = null;
    String reductionOp = null;
    boolean lastStatementInStatementList = false;

    public Spacing spc = new Spacing(3);   // see below for class Spacing

    public void nli(ReturnValue rv) { 
	((VectorReturnValue)rv).v.addElement("\n" + spc.spc);
    }
    public void din() { spc.updateSpc(+1); }
    public void dout() { spc.updateSpc(-1); }

    public boolean lexicallyInParallel() {
	return topmostParallelConstruct != null;
    }

    public void printCodeTo(PrintWriter pw) {
	this.rv.printTo(pw);
    }

    public void printExtraCodeTo(PrintWriter pw) {
	pw.print(extraCode());
    }

    public void printTypedefsTo(PrintWriter pw) {
	pw.print(typedefs);
    }

    public String extraCode() {
	StringBuffer retval = new StringBuffer();
	StringBuffer tmp = new StringBuffer();
	StringBuffer forAllOtherThreads = new StringBuffer();

	retval.append("void odinmp_thread_" + sourceFile.mangledName + "(odinmp_thread_data_t *odinmp_td)\n");
	retval.append("{\n");
	retval.append("  switch(" + OdinMP.THREAD_INFO("odinmp_td") + "->run_region) {\n");
	
	Enumeration e = parallelRegions.elements();
	while (e.hasMoreElements()) {
	    ParallelConstructDescriptor pcd = (ParallelConstructDescriptor)e.nextElement();
	    retval.append("  case " + pcd.index + ":\n");
	    retval.append(pcd.extraCode());
	    retval.append("    break;\n");
	}

	retval.append("  }\n");
	retval.append("}\n\n");

	e = threadprivateSymbols.symbolsInOrder.elements();
	while (e.hasMoreElements()) {
	    Symbol sym = (Symbol)e.nextElement();
	    if (sym.initializer != null) {
		sym.initializer.accept(this);
		String s = this.rv.toString();
		forAllOtherThreads.append("    " + sym + " = " + s + ";\n");
	    } else {
		forAllOtherThreads.append("    " + sym + ";\n");
	    }
	    tmp.append("    odinmp_td->threadprivate." + sym.name + " = &" + sym.name + ";\n");
	}
	retval.append("void odinmp_thread_init_" + sourceFile.mangledName + "(odinmp_thread_data_t *odinmp_td, int i) {\n");
	retval.append("  if (((odinmp_thread_info_t *)odinmp_td)->id == 0) {\n");
	retval.append(tmp);
	retval.append("    odinmp_thread_init(odinmp_td, i+1);\n");
	retval.append("  } else {\n");
	retval.append(forAllOtherThreads);
	retval.append(tmp);
	retval.append("    odinmp_thread_init(odinmp_td, i+1);\n");
	retval.append("  }\n\n");
	
	retval.append("}\n\n");

	if (mainInThisFile) {
	    IncrementalReader ir;
	    
	    try {
		ir = new IncrementalReader(new BufferedReader(new FileReader(new File(OdinMP.supporting(), "odinmp_main.c"))));
	    } catch (Exception ex) {
		System.err.println("Caught Exception trying to open incremental file: ``" + ex + "''");
		return null;
	    }

	    retval.append(ir.readIncrement());

	    if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
		if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		    retval.append("   if (odinmp_timing) {\n");
		}
		retval.append("  odinmp_timer_global_preStartTime = odinmp_time();\n");
		if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		    retval.append("   }\n");
		}
	    }

	    retval.append(ir.readIncrement());

	    if (OdinMP.targetTimingMode() != OdinMP.mode_Never) {
		if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		    retval.append("   if (odinmp_timing) {\n");
		}
		retval.append("  odinmp_timer_global_startTime = odinmp_time();\n");	
		if (OdinMP.targetTimingMode() == OdinMP.mode_Runtime) {
		    retval.append("   }\n");
		}
	    }

	    retval.append(ir.readIncrement());

	    try {
		ir.close();
	    } catch (Exception ex) {
		System.err.println("Caught Exception trying to close incremental file: ``" + ex + "''");
	    }	    
	}

	return new String(retval);
    }
    
    public Symbol symbolWithName(String name) {
	int nSymbolTables = symbolTableStack.size();
	int i;
	SymbolTable symbolTable;
	Symbol symbol = null;

	/* OdinMP.debugPrintln("Looking for symbol named ``" + name + "'' in SymbolTableStack with " + nSymbolTables + " entries"); */

	for (i = nSymbolTables - 1; i >= 0 && symbol == null; i--) {
	    symbolTable = (SymbolTable)symbolTableStack.elementAt(i);
	    symbol = symbolTable.symbolWithName(name);
	    /* OdinMP.debugPrintln("entry " + i + ": symbol = ``" + symbol + "''"); */
	}
	return symbol;
    }

    public SymbolTable symbolTableWithSymbolName(String name) {
	int nSymbolTables = symbolTableStack.size();
	int i;
	SymbolTable symbolTable = null;
	Symbol symbol = null;

	for (i = nSymbolTables - 1; i >= 0 && symbol == null; i--) {
	    symbolTable = (SymbolTable)symbolTableStack.elementAt(i);
	    symbol = symbolTable.symbolWithName(name);
	}
	return symbolTable;
    }

    public SymbolTable symbolTableWithSymbol(Symbol s) {
	int nSymbolTables = symbolTableStack.size();
	int i;
	SymbolTable symbolTable = null;

	for (i = nSymbolTables - 1; i >= 0; i--) {
	    symbolTable = (SymbolTable)symbolTableStack.elementAt(i);
	    if (symbolTable.containsSymbol(s))
		return symbolTable;
	}
	return null;
    }

    public boolean symbolIsGlobal(String name) {
	return symbolIsGlobal(symbolWithName(name));
    }

    public boolean symbolIsGlobal(Symbol sym) {
	return symbolTableWithSymbol(sym) == (SymbolTable)symbolTableStack.elementAt(0);
    }

    public void checkSharingForSymbol(Symbol sym) {
	int i;
	boolean hasDefinedSharing = false;
	for (i = constructDescriptorStack.size() - 1;
	     i >= 0 && !hasDefinedSharing;
	     i--) {
	    ConstructDescriptor cd = (ConstructDescriptor)constructDescriptorStack.elementAt(i);
	    
	    /* OdinMP.debugPrint("does `" + cd + "' contain symbol ``" + sym.name + "''?"); */
	    
	    if (cd.attributesForSymbols.containsKey(sym)) {
		OmpSymbolAttributes osa = (OmpSymbolAttributes)cd.attributesForSymbols.get(sym);
		/* OdinMP.debugPrintln(" Yes, it does."); */
		/* OdinMP.debugPrint("  So, does it set a sharing attribute? "); */
		
		if (osa.sharing != OmpSymbolAttributes.sharing_None) {
		    // we've found a sharing definition
		    hasDefinedSharing = true;
		    /* OdinMP.debugPrintln(" Yes, it does."); */
		} else {
		    /* OdinMP.debugPrintln(" No, it does not, let's search on."); */
		}
	    } else {
		/* OdinMP.debugPrintln(" No, it does not, let's search on."); */
	    }
	}
	    
	if (lexicallyInParallel()) {
	    if (!hasDefinedSharing) {
		if (defaultSharing == OmpSymbolAttributes.sharing_None) {
		    // a variable for which sharing can not be determined
		    System.err.println("sharing error: default sharing = none, but symbol ``" + sym.name + "'' sharing unspecified!");
		} else {
		    // default sharing, if not NONE, is 'shared'
		    ParallelConstructDescriptor pcd = topmostParallelConstruct;
		    OmpSymbolAttributes osa = pcd.attributesForSymbol(sym);
		    osa.setSharing(defaultSharing);
		    pcd.sharedSymbols.addSymbol(sym);
		    if (!symbolIsGlobal(sym))
			osa.nameReplacement = "(*odinmp_shared_data." + pcd + "." + sym.name + ")";
		    /* OdinMP.debugPrintln("symbol ``" + sym.name + "'' sharing set to " + defaultSharing); */
		}
	    }
	} else {
	    // If we're not lexically in a parallel region, then we really
	    // don't have to change anything!
	}
    }

    public ReturnValue replacedNameForIdentifier(Node n) {
	ReturnValue rv = newReturnValue();
	ReturnValue mrv = newReturnValue();
	n.accept(this); mrv.add(this.rv);
	String symName = lastAddedAsString(mrv);

	if (symName.equals("main")) {
	    ((VectorReturnValue)rv).v.addElement("odinmp_original_main");
	    mainInThisFile = true;
	    return rv;
	} 	
	
	Symbol sym = symbolWithName(symName);
	if (sym != null) {

	    checkSharingForSymbol(sym);
	    
	    String nameReplacement = nameReplacementForSymbol(sym);
	    if (OdinMP.debugMode)
		((VectorReturnValue)rv).v.addElement("/* " + symName + " -> */");
	    ((VectorReturnValue)rv).v.addElement(nameReplacement);
	} else {
	    rv = mrv;
	}
	return rv;
    }

    public String nameReplacementForSymbol(Symbol sym) {
	/* OdinMP.debugPrintln("looking for replacement name for ``" + sym.name + "''"); */

	if (sym.isThreadprivate) {
	    return "(*(odinmp_self_data()->threadprivate." + sym.name + "))";
	}

	for (int i = constructDescriptorStack.size() - 1;
	     i >= 0;
	     i--) {
	    ConstructDescriptor cd = (ConstructDescriptor)constructDescriptorStack.elementAt(i);

	    /* OdinMP.debugPrintln("  looking in `" + cd + "'"); */

	    if (cd.attributesForSymbols.containsKey(sym)) {
		/* OdinMP.debugPrintln("  `" + cd + "' has attributes for ``" + sym.name + "''"); */

		OmpSymbolAttributes osa = (OmpSymbolAttributes)cd.attributesForSymbols.get(sym);
		if (osa.sharing != OmpSymbolAttributes.sharing_None) {
		    /* OdinMP.debugPrintln("  `" + cd + "' has non-None = " + osa.sharing + " sharing attribute for ``" + sym.name + "''"); */
		    if (osa.nameReplacement != null) {
			/* OdinMP.debugPrintln("  `" + cd + "' has name replacement ``" + osa.nameReplacement + "'' for ``" + sym.name + "''"); */
			return osa.nameReplacement;
		    } else {
			/* OdinMP.debugPrintln("  `" + cd + "' has no name replacement for ``" + sym.name + "''"); */
			return sym.name;
		    }
		} else {
		    /* OdinMP.debugPrintln("  `" + cd + "' has a None sharing attribute for ``" + sym.name + "''"); */
		}
	    } else {
		/* OdinMP.debugPrintln("  `" + cd + "' has nothing for ``" + sym.name + "''"); */
	    }
	}
	return sym.name;
    }

    public ReturnValue newReturnValue() {
	return new VectorReturnValue();
    }

    public DataBuildingVisitor(OdinMP o) {
	odinmp = o;
	sourceFile = odinmp.currentSourceFile;
    }

    public String joinStringsInVector(Vector v){
	String s = "";
	Enumeration e = v.elements();
	if (e.hasMoreElements()) {
	    s = (String)e.nextElement(); 
	    while (e.hasMoreElements()) {
		s = s + sep + (String)e.nextElement();
	    }
	}
	return s;
    }

    public String lastAddedAsString(ReturnValue rv) {
	VectorReturnValue vrv = (VectorReturnValue)(rv.lastAdded);
	return joinStringsInVector(vrv.v);
    }

    //
    // Auto class visitors--probably don't need to be overridden.
    //
    public void visit(NodeList n) {
	ReturnValue rv = newReturnValue();
	for ( Enumeration e = n.elements(); e.hasMoreElements(); ) {
	    ((Node)e.nextElement()).accept(this); rv.add(this.rv);
	}
    	this.rv = rv;
    }

    public void visit(NodeListOptional n) {
	ReturnValue rv = newReturnValue();
	if ( n.present() )
	    for ( Enumeration e = n.elements(); e.hasMoreElements(); ) {
		((Node)e.nextElement()).accept(this); rv.add(this.rv);
	    }
    	this.rv = rv;
    }
    
    public void visit(NodeOptional n) {
	ReturnValue rv = newReturnValue();
	if ( n.present() ) {
	    n.node.accept(this); rv.add(this.rv);
	}
    	this.rv = rv;
    }
    
    public void visit(NodeSequence n) {
	ReturnValue rv = newReturnValue();
	for ( Enumeration e = n.elements(); e.hasMoreElements(); ) {
	    ((Node)e.nextElement()).accept(this); rv.add(this.rv);
	}
    	this.rv = rv;
    }
    
    public void visit(NodeToken n) {
	ReturnValue rv = newReturnValue(); 
	VectorReturnValue vrv = (VectorReturnValue)rv;
	VectorReturnValue svrv = new VectorReturnValue();

	int m = n.numSpecials();
	for (int i = 0; i < m; i++) {
	    svrv.v.addElement(n.getSpecialAt(i).toString());
	}

	vrv.v.addElement(n.tokenImage);
	
	/* OdinMP.debugPrintln("Consumed token [" + svrv + "] ``" + n.tokenImage + "''"); */
	
	this.rv = rv;
    }
    
    //
    // User-generated visitor methods below
    //

    /**
     * <PRE>
     * f0 -> ( ExternalDeclaration() | UnknownCpp() )+
     * </PRE>
     */
    public void visit(TranslationUnit n) {
	ReturnValue rv = newReturnValue();

	symbolTableStack.push(new SymbolTable());
	n.f0.accept(this); rv.add(this.rv);

	/* OdinMP.debugPrintln("TranslationUnit Symbol Table:\n--- begin ---\n" + symbolTableStack.peek() + "\n--- end --\n"); */
	
	symbolTableStack.pop();

    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ( FunctionDefinition() | Declaration() | ThreadprivateDirective() )
     * </PRE>
     */
    public void visit(ExternalDeclaration n) {
	ReturnValue rv = newReturnValue();

	n.f0.accept(this); rv.add(this.rv);
	nli(rv);

    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> [ DeclarationSpecifiers() ]
     * f1 -> Declarator()
     * f2 -> [ DeclarationList() ]
     * f3 -> CompoundStatement()
     * </PRE>
     */
    public void visit(FunctionDefinition n) {
	ReturnValue rv = newReturnValue();
	
	SymbolTable s = functionDefinitionSymbolTable = new SymbolTable();	

	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);

	symbolTableStack.push(functionDefinitionSymbolTable);
	functionDefinitionSymbolTable = null;

	n.f2.accept(this); rv.add(this.rv);
	nli(rv);

	/* OdinMP.debugPrintln("FunctionDefinition Symbol Table:\n--- begin ---\n" + s + "\n--- end --\n"); */

	n.f3.accept(this); rv.add(this.rv);
	nli(rv);

	symbolTableStack.pop();

    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> DeclarationSpecifiers()
     * f1 -> [ InitDeclaratorList() ]
     * f2 -> &lt;SEMICOLON&gt;
     * </PRE>
     */
    public void visit(Declaration n) {
	ReturnValue rv = newReturnValue();

	Decl d = new Decl();
	declStack.push(d);

	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
	n.f2.accept(this); rv.add(this.rv);

	/*
	  if (typedefSavedDeclStack != null) {
	  declStack = typedefSavedDeclStack;
	  typedefSavedDeclStack = null;
	  }
	*/
	d = (Decl)declStack.pop();
	if (d.declarationSpecifiers.contains("typedef")) {
	    typedefs.addSymbols(d.symbols);
	    rv = newReturnValue();
	} else {
	    ((SymbolTable)symbolTableStack.peek()).addSymbols(d.symbols);
	}

	if (lexicallyInParallel()) {
	    ParallelConstructDescriptor pcd = topmostParallelConstruct;
	    OmpSymbolAttributes osa;
	    Enumeration e;
	    Symbol sym;
	    if (d.declarationSpecifiers.contains("static")) {
		e = d.symbols.elements();
		while (e.hasMoreElements()) {
		    sym = (Symbol)e.nextElement();
		    osa = pcd.attributesForSymbol(sym);
		    osa.setSharing(OmpSymbolAttributes.sharing_Shared);
		}
	    } else {
		e = d.symbols.elements();
		while (e.hasMoreElements()) {
		    sym = (Symbol)e.nextElement();
		    osa = pcd.attributesForSymbol(sym);
		    osa.setSharing(OmpSymbolAttributes.sharing_Private);
		}
	    }
	}


	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ( Declaration() )+
     * </PRE>
     */
    public void visit(DeclarationList n) {
	ReturnValue rv = newReturnValue();
	Enumeration e = n.f0.elements();
	while (e.hasMoreElements()) {
	    Node nn = (Node)e.nextElement();
	    nn.accept(this); rv.add(this.rv);
	    nli(rv);
	}
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ( StorageClassSpecifier() | TypeSpecifier() | TypeQualifier() )+
     * </PRE>
     */
    public void visit(DeclarationSpecifiers n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;AUTO&gt;
     *       | &lt;REGISTER&gt;
     *       | &lt;STATIC&gt;
     *       | &lt;EXTERN&gt;
     *       | &lt;TYPEDEF&gt;
     * </PRE>
     */
    public void visit(StorageClassSpecifier n) {
	ReturnValue rv = newReturnValue();

	n.f0.accept(this); rv.add(this.rv);
	
	if (!declStack.empty()) {
	    ((Decl)declStack.peek()).add(lastAddedAsString(rv));
	    /*
	      if (n.f0.which == 4) { // it's a typedef
	      typedefSavedDeclStack = declStack;
	      declStack = new Stack();
	      }
	    */
	} 

    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;VOID&gt;
     *       | &lt;CHAR&gt;
     *       | &lt;SHORT&gt;
     *       | &lt;INT&gt;
     *       | &lt;LONG&gt;
     *       | &lt;FLOAT&gt;
     *       | &lt;DOUBLE&gt;
     *       | &lt;SIGNED&gt;
     *       | &lt;UNSIGNED&gt;
     *       | StructOrUnionSpecifier()
     *       | EnumSpecifier()
     *       | TypedefName()
     * </PRE>
     */
    public void visit(TypeSpecifier n) {
	ReturnValue rv = newReturnValue();

	n.f0.accept(this); rv.add(this.rv);
	String s = lastAddedAsString(rv);
	Symbol sym = typedefs.symbolWithName(s);

	/* OdinMP.debugPrintln(this + ": typedef symbol with name ``" + s + "'' = ``" + sym + "''"); */
	if (!declStack.empty()) {
	    ((Decl)declStack.peek()).add(s);
	    if (sym != null) {
		if (n.f0.which ==  9 || // StructOrUnionSpecifier
		    n.f0.which == 10 || // EnumSpecifier
		    n.f0.which == 11) { // TypedefName
		    ((Decl)declStack.peek()).dependOnType(sym);
		}
	    }
	}
	if (sym != null) {
	    if (savedDeclStack != null && !savedDeclStack.empty()) {
		if (n.f0.which ==  9 || // StructOrUnionSpecifier
		    n.f0.which == 10 || // EnumSpecifier
		    n.f0.which == 11) { // TypedefName
		    ((Decl)savedDeclStack.peek()).dependOnType(sym);
		}
	    }
	}

    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;CONST&gt;
     *       | &lt;VOLATILE&gt;
     * </PRE>
     */
    public void visit(TypeQualifier n) {
	ReturnValue rv = newReturnValue();

	    n.f0.accept(this); rv.add(this.rv);
	if (!declStack.empty()) {
	    ((Decl)declStack.peek()).add(lastAddedAsString(rv));
	}

    	this.rv = rv;
    }

    /**
     * <PRE>

     * f0 -> StructOrUnion()
     * f1 -> ( [ &lt;IDENTIFIER&gt; ] &lt;LEFTBRACE&gt; StructDeclarationList() &lt;RIGHTBRACE&gt; | &lt;IDENTIFIER&gt; )
     * </PRE>
     */
    public void visit(StructOrUnionSpecifier n) {
	ReturnValue rv = newReturnValue();

	Stack ds = null;
	boolean b = false;
	if (savedDeclStack == null) {
	    savedDeclStack = declStack;
	    b = true;
	} else {
	    ds = declStack;
	}
	declStack = new Stack();

	n.f0.accept(this); rv.add(this.rv);

	if (n.f1.which == 0) { // [ <IDENTIFIER> ] "{" StructDeclarationList() "}"
	    String structOrUnion = lastAddedAsString(rv);
	    NodeSequence ns = (NodeSequence) n.f1.choice;
	    Enumeration e = ns.elements();
	    VectorReturnValue vrv = new VectorReturnValue();

	    NodeOptional no = (NodeOptional)e.nextElement(); // [ <IDENTIFIER> ]
	    no.accept(this); vrv.add(this.rv);
	    String structOrUnionName = lastAddedAsString(vrv);

	    vrv = new VectorReturnValue();

	    Node nn = (Node)e.nextElement(); // "{"
	    nn.accept(this); vrv.add(this.rv);
	    din();
	    nli(vrv);

	    nn = (Node)e.nextElement(); // StructDeclarationList()
	    nn.accept(this); vrv.add(this.rv);
	    dout();
	    nli(vrv);

	    nn = (Node)e.nextElement(); // "}"
	    nn.accept(this); vrv.add(this.rv);

	    if (no.present()) {
		String structOrUnionDefined = vrv.toString();
		/* OdinMP.debugPrintln("StructOrUnionSpecifier: defining StructOrUnion ``" + structOrUnionDefined + "''"); */
		Symbol s = new Symbol();
		s.setName(structOrUnion + " " + structOrUnionName);
		s.add(structOrUnionDefined);
		typedefs.addSymbol(s);
		vrv = new VectorReturnValue();
		vrv.v.addElement(structOrUnionName);
		structOrUnionDefined = null;
	    }
	    rv.add(vrv);
	} else {
	    n.f1.accept(this); rv.add(this.rv);
	    String nam = rv.toString();
	    Symbol sym = typedefs.symbolWithName(nam);
	    if (sym == null) {
		Symbol s = new Symbol();
		s.setName(nam);
		typedefs.addSymbol(s);
	    }
	}

	if (b) {
	    declStack = savedDeclStack;
	    savedDeclStack = null;
	} else {
	    declStack = ds;
	}

    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;STRUCT&gt;
     *       | &lt;UNION&gt;
     * </PRE>
     */
    public void visit(StructOrUnion n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ( StructDeclaration() )+
     * </PRE>
     */
    public void visit(StructDeclarationList n) {
	ReturnValue rv = newReturnValue();

	Enumeration e = n.f0.elements();
	while (e.hasMoreElements()) {
	    Node nn = (Node)e.nextElement();
	    nn.accept(this); rv.add(this.rv);
	    if (e.hasMoreElements())
		nli(rv);
	}
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> InitDeclarator()
     * f1 -> ( &lt;COMMA&gt; InitDeclarator() )*
     * </PRE>
     */
    public void visit(InitDeclaratorList n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> Declarator()
     * f1 -> [ &lt;OP_ASS&gt; Initializer() ]
     * </PRE>
     */
    public void visit(InitDeclarator n) {
	ReturnValue rv = newReturnValue();

	if (!declStack.empty()) {
	    Decl d = (Decl)declStack.peek();
	    n.f0.accept(this); rv.add(this.rv);
	    if (n.f1.present()) {
		// Declarator just created a currentSymbol in d
		Symbol s = d.currentSymbol;
		n.f1.accept(this); rv.add(this.rv);
		NodeSequence ns = (NodeSequence)n.f1.node;
		s.initializer = (Initializer)ns.elementAt(1);;
	    }
	} else {
	    n.f0.accept(this); rv.add(this.rv);
	    n.f1.accept(this); rv.add(this.rv);
	}

    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> SpecifierQualifierList()
     * f1 -> StructDeclaratorList()
     * f2 -> &lt;SEMICOLON&gt;
     * </PRE>
     */
    public void visit(StructDeclaration n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
	n.f2.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ( TypeSpecifier() | TypeQualifier() )+
     * </PRE>
     */
    public void visit(SpecifierQualifierList n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> StructDeclarator()
     * f1 -> ( &lt;COMMA&gt; StructDeclarator() )*
     * </PRE>
     */
    public void visit(StructDeclaratorList n) {
	ReturnValue rv = newReturnValue();

	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);

    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> Declarator() [ &lt;COLON&gt; ConstantExpression() ]
     *       | &lt;COLON&gt; ConstantExpression()
     * </PRE>
     */
    public void visit(StructDeclarator n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;ENUM&gt;
     * f1 -> ( [ &lt;IDENTIFIER&gt; ] &lt;LEFTBRACE&gt; EnumeratorList() &lt;RIGHTBRACE&gt; | &lt;IDENTIFIER&gt; )
     * </PRE>
     */
    public void visit(EnumSpecifier n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	if (n.f1.which == 0) { // [ <IDENTIFIER> ] "{" EnumeratorList() "}"
	    String enum = lastAddedAsString(rv);
	    NodeSequence ns = (NodeSequence) n.f1.choice;
	    Enumeration e = ns.elements();
	    VectorReturnValue vrv = new VectorReturnValue();
	    NodeOptional no = (NodeOptional)e.nextElement(); // [ <IDENTIFIER> ]
	    no.accept(this); vrv.add(this.rv);
	    String enumName = lastAddedAsString(vrv);

	    vrv = new VectorReturnValue();

	    Node nn = (Node)e.nextElement(); // "{"
	    nn.accept(this); vrv.add(this.rv);
	    din(); nli(vrv);

	    nn = (Node)e.nextElement(); // EnumeratorList()
	    nn.accept(this); vrv.add(this.rv);
	    dout(); nli(vrv);

	    nn = (Node)e.nextElement(); // "}"
	    nn.accept(this); vrv.add(this.rv);

	    if (no.present()) {
		String enumDefined = vrv.toString();
		/* OdinMP.debugPrintln("EnumSpecifier: defining Enum ``" + enumDefined + "''"); */
		Symbol s = new Symbol();
		s.setName(enum + " " + enumName);
		s.add(enumDefined);
		typedefs.addSymbol(s);
		vrv = new VectorReturnValue();
		vrv.v.addElement(enumName);
		enumDefined = null;
	    }
	    rv.add(vrv);
	} else { // <IDENTIFIER>
	    n.f1.accept(this); rv.add(this.rv);
	}

    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> Enumerator()
     * f1 -> ( &lt;COMMA&gt; Enumerator() )*
     * </PRE>
     */
    public void visit(EnumeratorList n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;IDENTIFIER&gt;
     * f1 -> [ &lt;OP_ASS&gt; ConstantExpression() ]
     * </PRE>
     */
    public void visit(Enumerator n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> [ Pointer() ]
     * f1 -> DirectDeclarator()
     * </PRE>
     */
    public void visit(Declarator n) {
	ReturnValue rv = newReturnValue();
	boolean save = workingOnSymbol;

	if (!declStack.empty()) {
	    /* OdinMP.debugPrintln(this + ": Decl stack is non-empty"); */
	    if (!workingOnSymbol) {
		((Decl)declStack.peek()).newSymbol();
		/* OdinMP.debugPrintln("     created new Symbol `" + ((Decl)declStack.peek()).currentSymbol + "' in Decl `" + (Decl)declStack.peek() + "'"); */
		workingOnSymbol = true;
	    } else {
		/* OdinMP.debugPrintln("     already working on Symbol `" + ((Decl)declStack.peek()).currentSymbol + "'"); */
	    }
	} else {
	    /* OdinMP.debugPrintln(this + ": Decl stack is empty, doing nothin"); */
	}
	
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
	workingOnSymbol = save;

    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ( &lt;IDENTIFIER&gt; | &lt;LEFTPAREN&gt; Declarator() &lt;RIGHTPAREN&gt; )
     * f1 -> ( &lt;LEFTBRACKET&gt; [ ConstantExpression() ] &lt;RIGHTBRACKET&gt; | &lt;LEFTPAREN&gt; ParameterTypeList() &lt;RIGHTPAREN&gt; | &lt;LEFTPAREN&gt; [ IdentifierList() ] &lt;RIGHTPAREN&gt; )*
     * </PRE>
     */
    public void visit(DirectDeclarator n) {
	ReturnValue rv = newReturnValue();

	if (!declStack.empty()) {
	    Decl d = (Decl)declStack.peek();
	    if (n.f0.which == 0) { // it's an identifier
		n.f0.accept(this); rv.add(this.rv);
		d.currentSymbol.setName(lastAddedAsString(rv));
	    } else { // it's a declarator in parentheses
		d.add("(");
		n.f0.accept(this); rv.add(this.rv);
		d.add(")");
	    }

	    n.f1.accept(this); rv.add(this.rv);
	    d.add(lastAddedAsString(rv));
	} else if (functionDefinitionSymbolTable != null) {
	    if (n.f0.which == 0) { // it's an identifier
		n.f0.accept(this); rv.add(this.rv);
		String s = lastAddedAsString(rv);
		if (s.equals("main")) { // we're defining main!
		    rv = newReturnValue();
		    ((VectorReturnValue)rv).v.addElement("odinmp_original_main");
		    mainInThisFile = true;
		}
	    } else {
		n.f0.accept(this); rv.add(this.rv);
	    }
	    n.f1.accept(this); rv.add(this.rv);
	} else {
	    n.f0.accept(this); rv.add(this.rv);
	    n.f1.accept(this); rv.add(this.rv);
	}

    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;OP_MUL&gt;
     * f1 -> [ TypeQualifierList() ]
     * f2 -> [ Pointer() ]
     * </PRE>
     */
    public void visit(Pointer n) {
	ReturnValue rv = newReturnValue();

	if (!declStack.empty()) {
	    Decl d = (Decl)declStack.peek();
	    d.add("*");
	}
	
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
	n.f2.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ( TypeQualifier() )+
     * </PRE>
     */
    public void visit(TypeQualifierList n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ParameterList()
     * f1 -> [ &lt;COMMA&gt; &lt;ELLIPSIS&gt; ]
     * </PRE>
     */
    public void visit(ParameterTypeList n) {
	ReturnValue rv = newReturnValue();

	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);

    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ParameterDeclaration()
     * f1 -> ( &lt;COMMA&gt; ParameterDeclaration() )*
     * </PRE>
     */
    public void visit(ParameterList n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> DeclarationSpecifiers()
     * f1 -> ( Declarator() | [ AbstractDeclarator() ] )
     * </PRE>
     */
    public void visit(ParameterDeclaration n) {
	ReturnValue rv = newReturnValue();

	Decl d;
	declStack.push(d = new Decl());
	boolean save = workingOnSymbol;
	workingOnSymbol = false;
	
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);

	if (functionDefinitionSymbolTable != null) {
	    functionDefinitionSymbolTable.addSymbols(d.symbols);
	}
	declStack.pop();
	workingOnSymbol = save;

    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;IDENTIFIER&gt;
     * f1 -> ( &lt;COMMA&gt; &lt;IDENTIFIER&gt; )*
     * </PRE>
     */
    public void visit(IdentifierList n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> AssignmentExpression()
     *       | &lt;LEFTBRACE&gt; InitializerList() [ &lt;COMMA&gt; ] &lt;RIGHTBRACE&gt;
     * </PRE>
     */
    public void visit(Initializer n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> Initializer()
     * f1 -> ( &lt;COMMA&gt; Initializer() )*
     * </PRE>
     */
    public void visit(InitializerList n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> SpecifierQualifierList()
     * f1 -> [ AbstractDeclarator() ]
     * </PRE>
     */
    public void visit(TypeName n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ( Pointer() | [ Pointer() ] DirectAbstractDeclarator() )
     * </PRE>
     */
    public void visit(AbstractDeclarator n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ( &lt;LEFTPAREN&gt; AbstractDeclarator() &lt;RIGHTPAREN&gt; | &lt;LEFTBRACKET&gt; [ ConstantExpression() ] &lt;RIGHTBRACKET&gt; | &lt;LEFTPAREN&gt; [ ParameterTypeList() ] &lt;RIGHTPAREN&gt; )
     * f1 -> ( &lt;LEFTBRACKET&gt; [ ConstantExpression() ] &lt;RIGHTBRACKET&gt; | &lt;LEFTPAREN&gt; [ ParameterTypeList() ] &lt;RIGHTPAREN&gt; )*
     * </PRE>
     */
    public void visit(DirectAbstractDeclarator n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;IDENTIFIER&gt;
     * </PRE>
     */
    public void visit(TypedefName n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ( LabeledStatement() | ExpressionStatement() | CompoundStatement() | SelectionStatement() | IterationStatement() | JumpStatement() | OmpConstruct() | OmpDirective() | UnknownCpp() )
     * </PRE>
     */
    public void visit(Statement n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;CROSSBAR&gt;
     * f1 -> &lt;UNKNOWN_CPP&gt;
     * </PRE>
     */
    public void visit(UnknownCpp n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
	((VectorReturnValue)rv).v.addElement("\n");
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;OMP_CR&gt; | &lt;OMP_NR&gt;
     * </PRE>
     */
    public void visit(OmpEol n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> OmpPragma()
     * f1 -> ( ParallelConstruct() | ParallelForConstruct() | ParallelSectionsConstruct() | ForConstruct() | SectionsConstruct() | SingleConstruct() | MasterConstruct() | CriticalConstruct() | AtomicConstruct() | OrderedConstruct() )
     * </PRE>
     */
    public void visit(OmpConstruct n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> OmpPragma()
     * f1 -> ( BarrierDirective() | FlushDirective() )
     * </PRE>
     */
    public void visit(OmpDirective n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ParallelDirective()
     * f1 -> CompoundStatement()
     * </PRE>
     */
    public void visit(ParallelConstruct n) {
	ReturnValue rv = newReturnValue();
	ParallelConstructDescriptor savedPcd = topmostParallelConstruct;

	ParallelConstructDescriptor pcd = topmostParallelConstruct = new ParallelConstructDescriptor(n.f1);

	clauseHandler = new ParallelClauseHandler(this, pcd);
	n.f0.accept(this); // don't add it! // rv.add(this.rv);
	clauseHandler = null;
	nli(rv);

	VectorReturnValue vrv = (VectorReturnValue)rv;
	vrv.v.addElement(pcd.code(this));
	parallelRegions.addElement(pcd);
	
	topmostParallelConstruct = savedPcd;

    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;CROSSBAR&gt;
     * f1 -> &lt;PRAGMA&gt;
     * f2 -> &lt;OMP&gt;
     * </PRE>
     */
    public void visit(OmpPragma n) {
	ReturnValue rv = newReturnValue();

	((VectorReturnValue)rv).v.addElement("/* ");
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
	n.f2.accept(this); rv.add(this.rv);
	((VectorReturnValue)rv).v.addElement(" */");

	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;CROSSBAR&gt;
     * f1 -> &lt;PRAGMA&gt;
     * f2 -> &lt;UNKNOWN_CPP&gt;
     * </PRE>
     */
    public void visit(UnknownPragma n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this);
	n.f1.accept(this);
	n.f2.accept(this);
	((VectorReturnValue)rv).v.addElement("\n");
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;PARALLEL&gt;
     * f1 -> ( UniqueParallelClause() | DataClause() )*
     * f2 -> OmpEol()
     * </PRE>
     */
    public void visit(ParallelDirective n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
	n.f2.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;IF&gt;
     * f1 -> &lt;LEFTPAREN&gt;
     * f2 -> Expression()
     * f3 -> &lt;RIGHTPAREN&gt;
     * </PRE>
     */
    public void visit(UniqueParallelClause n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
	n.f2.accept(this); rv.add(this.rv);
	n.f3.accept(this); rv.add(this.rv);

	clauseHandler.handleIf(n.f2);

    	this.rv = rv;
    }

   /**
    * <PRE>
    * f0 -> ForDirective()
    * f1 -> OmpForHeader()
    * f2 -> Statement()
    * </PRE>
    */
    public void visit(ForConstruct n) {
	ReturnValue rv = newReturnValue();
	VectorReturnValue vrv = (VectorReturnValue)rv;

	ConstructDescriptor savedWscd = topmostWorkSharingConstruct;
	ForConstructDescriptor fcd = new ForConstructDescriptor(n.f2);
	topmostWorkSharingConstruct = fcd;
	clauseHandler = new ForClauseHandler(this, fcd);
	n.f0.accept(this); // rv.add(this.rv);
	nli(rv);
	clauseHandler = null;
	n.f1.accept(this); // rv.add(this.rv);
	// n.f2.accept(this); // rv.add(this.rv);

	vrv.v.addElement(fcd.code(this));
	topmostWorkSharingConstruct = savedWscd;

	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;FOR&gt;
     * f1 -> ( UniqueForClause() | DataClause() | &lt;NOWAIT&gt; )*
     * f2 -> OmpEol();
     * </PRE>
     */
    public void visit(ForDirective n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
	n.f2.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;ORDERED&gt;
     *       | &lt;SCHEDULE&gt; &lt;LEFTPAREN&gt; ScheduleKind() [ &lt;COMMA&gt; Expression() ] &lt;RIGHTPAREN&gt;
     * </PRE>
     */
    public void visit(UniqueForClause n) {
	ReturnValue rv = newReturnValue();
	if (n.f0.which == 0) { // <ORDERED>
	    n.f0.accept(this); rv.add(this.rv);
	    clauseHandler.handleOrdered();
	} else {
	    NodeSequence ns = (NodeSequence)n.f0.choice;
	    Enumeration e = ns.elements();
	    Expression chunkSize = null;

	    ((Node)e.nextElement()).accept(this); rv.add(this.rv); // <SCHEDULE>
	    ((Node)e.nextElement()).accept(this); rv.add(this.rv); // "("
	    ((Node)e.nextElement()).accept(this); rv.add(this.rv); // ScheduleKind()
	    String schedulerName = lastAddedAsString(rv);

	    NodeOptional no = (NodeOptional)e.nextElement();
	    if (no.present()) {
		NodeSequence ns2 = (NodeSequence)no.node;
		Enumeration e2 = ns2.elements();
		((Node)e2.nextElement()).accept(this); rv.add(this.rv); // ","
		chunkSize = ((Expression)e2.nextElement()); // Expression()
		chunkSize.accept(this); rv.add(this.rv); // Expression
	    }
	    ((Node)e.nextElement()).accept(this); rv.add(this.rv); // ")"
	    
	    clauseHandler.handleSchedule(schedulerName, chunkSize);
	}
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;STATIC&gt;
     *       | &lt;DYNAMIC&gt;
     *       | &lt;GUIDED&gt;
     *       | &lt;RUNTIME&gt;
     * </PRE>
     */
    public void visit(ScheduleKind n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    
    /**
     * <PRE>
     * f0 -> &lt;FOR&gt;
     * f1 -> &lt;LEFTPAREN&gt;
     * f2 -> OmpForInitExpression()
     * f3 -> &lt;SEMICOLON&gt;
     * f4 -> OmpForCondition()
     * f5 -> &lt;SEMICOLON&gt;
     * f6 -> OmpForReinitExpression()
     * f7 -> &lt;RIGHTPAREN&gt;
     * </PRE>
     */
    public void visit(OmpForHeader n) {
	n.f2.accept(this);
	n.f4.accept(this); 
	n.f6.accept(this);
	this.rv = newReturnValue();
    }

    /**
     * <PRE>
     * f0 -> &lt;IDENTIFIER&gt;
     * f1 -> &lt;OP_ASS&gt;
     * f2 -> Expression()
     * </PRE>
     */
    public void visit(OmpForInitExpression n) {
	ForConstructDescriptor fcd = (ForConstructDescriptor)topmostWorkSharingConstruct;
	fcd.loopIndex = symbolWithName(n.f0.tokenImage);
	fcd.initExpression = n.f2;
	this.rv = newReturnValue();
    }

    /**
     * <PRE>
     * f0 -> &lt;IDENTIFIER&gt;
     * f1 -> ( &lt;OP_LT&gt; | &lt;OP_LE&gt; | &lt;OP_GT&gt; | &lt;OP_GE&gt; )
     * f2 -> Expression()
     * </PRE>
     */
    public void visit(OmpForCondition n) {
	ForConstructDescriptor fcd = (ForConstructDescriptor)topmostWorkSharingConstruct;
	switch (n.f1.which) {
	case 0: // <OP_LT>
	    fcd.comparisonOperator =
		ForConstructDescriptor.comparisonOperator_LT;
	    break;
	case 1: // <OP_LE>
	    fcd.comparisonOperator =
		ForConstructDescriptor.comparisonOperator_LE;
	    break;
	case 2: // <OP_GT>
	    fcd.comparisonOperator =
		ForConstructDescriptor.comparisonOperator_GT;
	    break;
	case 3: // <OP_GE>
	    fcd.comparisonOperator =
		ForConstructDescriptor.comparisonOperator_GE;
	    break;
	default: // huh? impossible!
	}
	fcd.compareAgainst = n.f2;
	this.rv = newReturnValue();
    }

    /**
     * <PRE>
     * f0 -> ( &lt;OP_INCR&gt; | &lt;OP_DECR&gt; ) &lt;IDENTIFIER&gt;
     *       | &lt;IDENTIFIER&gt; ( &lt;OP_INCR&gt; | &lt;OP_DECR&gt; )
     *       | &lt;IDENTIFIER&gt; ( &lt;OP_ADDASS&gt; | &lt;OP_SUBASS&gt; ) Expression()
     *       | &lt;IDENTIFIER&gt; &lt;OP_ASS&gt; ( &lt;IDENTIFIER&gt; ( &lt;OP_ADD&gt; | &lt;OP_SUB&gt; ) AdditiveExpression() | MultiplicativeExpression() &lt;OP_ADD&gt; &lt;IDENTIFIER&gt; )
     * </PRE>
     */
    public void visit(OmpForReinitExpression n) {
	NodeSequence ns = (NodeSequence)n.f0.choice;
	NodeChoice nc;
	ForConstructDescriptor fcd = (ForConstructDescriptor)topmostWorkSharingConstruct;

	switch (n.f0.which) {
	case 0: // ++var | --var
	    nc = (NodeChoice)ns.elementAt(0);
	    if (nc.which == 0)
		fcd.setIncrement(1);
	    else
		fcd.setIncrement(-1);
	    
	    break;
	case 1: // var++ | var--
	    nc = (NodeChoice)ns.elementAt(1);
	    if (nc.which == 0)
		fcd.setIncrement(1);
	    else
		fcd.setIncrement(-1);
	    
	    break;
	case 2: // var += Expression() | var -= Expression()
	    Expression exp = (Expression)ns.elementAt(2);

	    nc = (NodeChoice)ns.elementAt(1);
	    if (nc.which == 0)
		fcd.setIncrement(1, exp);
	    else
		fcd.setIncrement(-1, exp);
	    
	    break;
	case 3: // var = ...
	    nc = (NodeChoice)ns.elementAt(2);
	    ns = (NodeSequence)nc.choice;
	    switch(nc.which) {
	    case 0: // var + Expression() | var - Expression()
		AdditiveExpression aexp = (AdditiveExpression)ns.elementAt(2);
		
		nc = (NodeChoice)ns.elementAt(1);
		if (nc.which == 0)
		    fcd.setIncrement(1, aexp);
		else
		    fcd.setIncrement(-1, aexp);

		break;
	    case 1: // Expression() + var
		MultiplicativeExpression mexp = (MultiplicativeExpression)ns.elementAt(0);
		
		fcd.setIncrement(1, mexp);

		break;
	    default: // Huh? Not possible!
		break;
	    }
	    break;
	default: // Huh? Not possible!
	    break;
	}
	this.rv = newReturnValue();
    }
    
    /**
     * <PRE>
     * f0 -> &lt;SECTIONS&gt;
     * f1 -> ( &lt;NOWAIT&gt; | DataClause() )*
     * f2 -> OmpEol();
     * f3 -> SectionsScope()
     * </PRE>
     */
    public void visit(SectionsConstruct n) {
	ReturnValue rv = newReturnValue();
	VectorReturnValue vrv = (VectorReturnValue)rv;

	ConstructDescriptor savedWscd = topmostWorkSharingConstruct;
	SectionsConstructDescriptor scd = new SectionsConstructDescriptor();
	topmostWorkSharingConstruct = scd;
	clauseHandler = new SectionsClauseHandler(this, scd);
	// n.f0.accept(this); rv.add(this.rv);
	if (n.f1.present()) {
	    Enumeration e = n.f1.elements();
	    while (e.hasMoreElements()) {
		NodeChoice n2 = (NodeChoice)e.nextElement();
		n2.accept(this); // rv.add(this.rv);
		if (n2.which == 0) { // <NOWAIT>
		    clauseHandler.handleNowait();
		} 
	    }
	}
	n.f2.accept(this); rv.add(this.rv);
	clauseHandler = null;
	n.f3.accept(this); 

	vrv.v.addElement(scd.code(this));
	topmostWorkSharingConstruct = savedWscd;

	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;LEFTBRACE&gt;
     * f1 -> [ Statement() ]
     * f2 -> ( OmpPragma() &lt;SECTION&gt; OmpEol() Statement() )*
     * f3 -> &lt;RIGHTBRACE&gt;
     * </PRE>
     */
    public void visit(SectionsScope n) {
	SectionsConstructDescriptor scd = (SectionsConstructDescriptor)topmostWorkSharingConstruct;
	// n.f0.accept(this); rv.add(this.rv);
	if (n.f1.present()) {
	    scd.sections.addElement(n.f1.node);
	}
	Enumeration e = n.f2.elements();
	while (e.hasMoreElements()) {
	    NodeSequence ns = (NodeSequence)e.nextElement();
	    scd.sections.addElement((Statement)ns.elementAt(3));
	}
	// n.f3.accept(this); rv.add(this.rv);
    	this.rv = newReturnValue();
    }

    /**
     * <PRE>
     * f0 -> &lt;SINGLE&gt;
     * f1 -> ( &lt;NOWAIT&gt; | DataClause() )*
     * f2 -> OmpEol()
     * f3 -> Statement()
     * </PRE>
     */
    public void visit(SingleConstruct n) {
	ReturnValue rv = newReturnValue();
	VectorReturnValue vrv = (VectorReturnValue)rv;

	ConstructDescriptor savedWscd = topmostWorkSharingConstruct;
	SingleConstructDescriptor scd = new SingleConstructDescriptor(n.f3);
	topmostWorkSharingConstruct = scd;

	clauseHandler = new SingleClauseHandler(this, scd);

	// n.f0.accept(this); rv.add(this.rv);
	if (n.f1.present()) {
	    Enumeration e = n.f1.elements();
	    while (e.hasMoreElements()) {
		NodeChoice n2 = (NodeChoice)e.nextElement();
		if (n2.which == 0) { // <NOWAIT>
		    clauseHandler.handleNowait();
		} else {
		    n2.accept(this); // rv.add(this.rv);
		}
	    }
	}
	// n.f2.accept(this); rv.add(this.rv);
	clauseHandler = null;

	vrv.v.addElement(scd.code(this));
	topmostWorkSharingConstruct = savedWscd;
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;PARALLEL&gt;
     * f1 -> &lt;FOR&gt;
     * f2 -> ( UniqueParallelClause() | UniqueForClause() | DataClause() )*
     * f3 -> OmpEol()
     * f4 -> OmpForHeader()
     * f5 -> Statement()
     * </PRE>
     */
    public void visit(ParallelForConstruct n) {
	ReturnValue rv = newReturnValue();
	VectorReturnValue vrv = (VectorReturnValue)rv;
	
	ConstructDescriptor savedWscd = topmostWorkSharingConstruct;
	ForConstructDescriptor fcd = new ForConstructDescriptor(n.f5);
	topmostWorkSharingConstruct = fcd;
	fcd.nowait = true;

	ParallelConstructDescriptor savedPcd = topmostParallelConstruct;
	ParallelConstructDescriptor pcd = topmostParallelConstruct = new ParallelConstructDescriptor(fcd);

	clauseHandler = new ParallelForClauseHandler(this, pcd, fcd);
	// n.f0.accept(this); rv.add(this.rv);
	// n.f1.accept(this); rv.add(this.rv);
	n.f2.accept(this); // rv.add(this.rv);
	clauseHandler = null;
	constructDescriptorStack.push(pcd);
	n.f4.accept(this); // rv.add(this.rv);
	constructDescriptorStack.pop();
	// n.f4.accept(this); rv.add(this.rv); // don't handle this one!

	vrv.v.addElement(pcd.code(this));
	parallelRegions.addElement(pcd);

	topmostWorkSharingConstruct = savedWscd;
	topmostParallelConstruct = savedPcd;
	
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;PARALLEL&gt;
     * f1 -> &lt;SECTIONS&gt;
     * f2 -> ( UniqueParallelClause() | DataClause() )*
     * f3 -> OmpEol()
     * f4 -> SectionsScope()
     * </PRE>
     */
    public void visit(ParallelSectionsConstruct n) {
	ReturnValue rv = newReturnValue();
	VectorReturnValue vrv = (VectorReturnValue)rv;

	ConstructDescriptor savedWscd = topmostWorkSharingConstruct;
	SectionsConstructDescriptor scd = new SectionsConstructDescriptor();
	topmostWorkSharingConstruct = scd;
	scd.nowait = true;

	ParallelConstructDescriptor savedPcd = topmostParallelConstruct;
	ParallelConstructDescriptor pcd = topmostParallelConstruct = new ParallelConstructDescriptor(scd);
	
	clauseHandler = new ParallelSectionsClauseHandler(this, pcd, scd);
	// n.f0.accept(this); rv.add(this.rv);
	// n.f1.accept(this); rv.add(this.rv);
	n.f2.accept(this); // rv.add(this.rv);
	clauseHandler = null;
	constructDescriptorStack.push(pcd);
	n.f4.accept(this); // rv.add(this.rv);
	constructDescriptorStack.pop();

	vrv.v.addElement(pcd.code(this));
	parallelRegions.addElement(pcd);

	topmostWorkSharingConstruct = savedWscd;
	topmostParallelConstruct = savedPcd;
	
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;MASTER&gt;
     * f1 -> OmpEol()
     * f2 -> CompoundStatement()
     * </PRE>
     */
    public void visit(MasterConstruct n) {
	ReturnValue rv = newReturnValue();
	VectorReturnValue vrv = (VectorReturnValue)rv;
	//n.f0.accept(this); rv.add(this.rv);
	//n.f1.accept(this); rv.add(this.rv);

	vrv.v.addElement(new MasterConstructDescriptor(n.f2).code(this));
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;CRITICAL&gt;
     * f1 -> [ &lt;LEFTPAREN&gt; &lt;IDENTIFIER&gt; &lt;RIGHTPAREN&gt; ]
     * f2 -> OmpEol()
     * f3 -> Statement()
     * </PRE>
     */
    public void visit(CriticalConstruct n) {
	ReturnValue rv = newReturnValue();
	VectorReturnValue vrv = (VectorReturnValue)rv;

	CriticalConstructDescriptor ccd = new CriticalConstructDescriptor(n.f3);
	// n.f0.accept(this); rv.add(this.rv);
	if (n.f1.present()) {
	    NodeSequence ns = (NodeSequence)n.f1.node;
	    ccd.name = ((NodeToken)ns.elementAt(1)).tokenImage;
	}

	vrv.v.addElement(ccd.code(this));
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;ATOMIC&gt;
     * f1 -> OmpEol()
     * f2 -> ExpressionStatement()
     * </PRE>
     */
    public void visit(AtomicConstruct n) {
	ReturnValue rv = newReturnValue();
	VectorReturnValue vrv = (VectorReturnValue)rv;

	vrv.v.addElement(new AtomicConstructDescriptor(n.f2).code(this));

    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;FLUSH&gt;
     * f1 -> [ &lt;LEFTPAREN&gt; VariableList() &lt;RIGHTPAREN&gt; ]
     * f2 -> OmpEol()
     * </PRE>
     */
    public void visit(FlushDirective n) {
	ReturnValue rv = newReturnValue();
	VectorReturnValue vrv = (VectorReturnValue)rv;

	variableList = null;
	// n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); // rv.add(this.rv);

	vrv.v.addElement(new FlushConstructDescriptor(variableList).code(this));
	variableList = null;
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;ORDERED&gt;
     * f1 -> OmpEol()
     * f2 -> CompoundStatement()
     * </PRE>
     */
    public void visit(OrderedConstruct n) {
	ReturnValue rv = newReturnValue();
	VectorReturnValue vrv = (VectorReturnValue)rv;
	OrderedConstructDescriptor ocd = new OrderedConstructDescriptor(n.f2);
	vrv.v.addElement(ocd.code(this));
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;BARRIER&gt;
     * f1 -> OmpEol()
     * </PRE>
     */
    public void visit(BarrierDirective n) {
	ReturnValue rv = newReturnValue();
	VectorReturnValue vrv = (VectorReturnValue)rv;

	BarrierConstructDescriptor bcd = new BarrierConstructDescriptor();

	vrv.v.addElement(bcd.code(this));
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> OmpPragma()
     * f1 -> &lt;THREADPRIVATE&gt;
     * f2 -> &lt;LEFTPAREN&gt;
     * f3 -> VariableList()
     * f4 -> &lt;RIGHTPAREN&gt;
     * f5 -> OmpEol()
     * </PRE>
     */
    public void visit(ThreadprivateDirective n) {

	// n.f0.accept(this); rv.add(this.rv);
	// n.f1.accept(this); rv.add(this.rv);
	// n.f2.accept(this); rv.add(this.rv);

	n.f3.accept(this); // rv.add(this.rv);
	Enumeration e = variableList.elements();
	while (e.hasMoreElements()) {
	    String nam = (String)e.nextElement();
	    Symbol sym = symbolWithName(nam);
	    if (!sym.declarationSpecifiers.contains("extern")) {
		// if it's an extern symbol, then we don't add it to the
		// list of threadprivate symbols
		threadprivateSymbols.addSymbol(sym);
		odinmp.threadprivateSymbols.addSymbol(sym);
	    }
	    sym.isThreadprivate = true;
	}

	// n.f4.accept(this); rv.add(this.rv);

    	this.rv = newReturnValue();
    }

    /**
     * <PRE>
     * f0 -> &lt;PRIVATE&gt; &lt;LEFTPAREN&gt; VariableList() &lt;RIGHTPAREN&gt;
     *       | &lt;FIRSTPRIVATE&gt; &lt;LEFTPAREN&gt; VariableList() &lt;RIGHTPAREN&gt;
     *       | &lt;LASTPRIVATE&gt; &lt;LEFTPAREN&gt; VariableList() &lt;RIGHTPAREN&gt;
     *       | &lt;SHARED&gt; &lt;LEFTPAREN&gt; VariableList() &lt;RIGHTPAREN&gt;
     *       | &lt;COPYIN&gt; &lt;LEFTPAREN&gt; VariableList() &lt;RIGHTPAREN&gt;
     *       | &lt;DFLT&gt; &lt;LEFTPAREN&gt; ( &lt;SHARED&gt; | &lt;NONE&gt; ) &lt;RIGHTPAREN&gt;
     *       | &lt;REDUCTION&gt; &lt;LEFTPAREN&gt; ReductionOp() &lt;COLON&gt; VariableList() &lt;RIGHTPAREN&gt;
     * </PRE>
     */
    public void visit(DataClause n) {
	ReturnValue rv = newReturnValue();

	NodeSequence ns = (NodeSequence)n.f0.choice;
	Enumeration e = ns.elements();
	String s;
	Hashtable h;

	switch (n.f0.which) {
	case 0: // <PRIVATE>
	case 1: // <FIRSTPRIVATE>
	case 2: // <LASTPRIVATE>
	case 3: // <SHARED>
	case 4: // <COPYIN>
	    variableList = null;
	    ((Node)e.nextElement()).accept(this); rv.add(this.rv); // the 'tag'
	    s = lastAddedAsString(rv);
	    ((Node)e.nextElement()).accept(this); rv.add(this.rv); // "("
	    ((Node)e.nextElement()).accept(this); rv.add(this.rv); // VariableList()
	    ((Node)e.nextElement()).accept(this); rv.add(this.rv); // ")"
	    switch(n.f0.which) {
	    case 0: clauseHandler.handlePrivate(variableList); break;
	    case 1: clauseHandler.handleFirstprivate(variableList); break;
	    case 2: clauseHandler.handleLastprivate(variableList); break;
	    case 3: clauseHandler.handleShared(variableList); break;
	    case 4: clauseHandler.handleCopyin(variableList); break;
	    }
	    break;
	case 5: // <DFLT>
	    ((Node)e.nextElement()).accept(this); rv.add(this.rv); // <DFLT>
	    ((Node)e.nextElement()).accept(this); rv.add(this.rv); // "("
	    ((Node)e.nextElement()).accept(this); rv.add(this.rv); // <SHARED>|<NONE>
	    s = lastAddedAsString(rv);
	    ((Node)e.nextElement()).accept(this); rv.add(this.rv); // ")"
	    clauseHandler.handleDefault(s);
	    break;
	case 6: // <REDUCTION>
	    ((Node)e.nextElement()).accept(this); rv.add(this.rv); // <REDUCTION>
	    ((Node)e.nextElement()).accept(this); rv.add(this.rv); // "("
	    ((Node)e.nextElement()).accept(this); rv.add(this.rv); // ReductionOp;
	    s = lastAddedAsString(rv);
	    ((Node)e.nextElement()).accept(this); rv.add(this.rv); // ":"
	    ((Node)e.nextElement()).accept(this); rv.add(this.rv); // VariableList()
	    ((Node)e.nextElement()).accept(this); rv.add(this.rv); // ")"
	    clauseHandler.handleReduction(s, variableList);

	    break;
	default: // should not ever happen!
	    System.err.println("What? an unknown data clause???");
	    break;
	}

    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;OP_ADD&gt;
     *       | &lt;OP_MUL&gt;
     *       | &lt;OP_SUB&gt;
     *       | &lt;OP_BITAND&gt;
     *       | &lt;OP_BITXOR&gt;
     *       | &lt;OP_BITOR&gt;
     *       | &lt;OP_OR&gt;
     *       | &lt;OP_AND&gt;
     * </PRE>
     */
    public void visit(ReductionOp n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	reductionOp = lastAddedAsString(rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;IDENTIFIER&gt;
     * f1 -> ( &lt;COMMA&gt; &lt;IDENTIFIER&gt; )*
     * </PRE>
     */
    public void visit(VariableList n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);

	variableList = new Vector();
	Enumeration e = ((VectorReturnValue)rv).v.elements();
	variableList.addElement(e.nextElement());
	while (e.hasMoreElements()) {
	    e.nextElement();
	    variableList.addElement(e.nextElement());
	}
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ( &lt;IDENTIFIER&gt; &lt;COLON&gt; Statement() | &lt;CASE&gt; ConstantExpression() &lt;COLON&gt; Statement() | &lt;DFLT&gt; &lt;COLON&gt; Statement() )
     * </PRE>
     */
    public void visit(LabeledStatement n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> [ Expression() ]
     * f1 -> &lt;SEMICOLON&gt;
     * </PRE>
     */
    public void visit(ExpressionStatement n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;LEFTBRACE&gt;
     * f1 -> [ DeclarationList() ]
     * f2 -> [ StatementList() ]
     * f3 -> &lt;RIGHTBRACE&gt;
     * </PRE>
     */
    public void visit(CompoundStatement n) {
	ReturnValue rv = newReturnValue();

	symbolTableStack.push(new SymbolTable());

	n.f0.accept(this); rv.add(this.rv);
	din();
	if (n.f1.present()) {
	    nli(rv);
	    n.f1.accept(this); rv.add(this.rv);
	    if (n.f2.present()) {
		nli(rv);
	    }
	}
	if (n.f2.present()) {
	    nli(rv);
	    n.f2.accept(this); rv.add(this.rv);
	}
	dout();
	nli(rv);

	/* OdinMP.debugPrintln("CompoundStatement Symbol Table:\n--- begin ---\n" + symbolTableStack.peek() + "\n--- end --\n"); */

	n.f3.accept(this); rv.add(this.rv);

	symbolTableStack.pop();

    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ( Statement() )+
     * </PRE>
     */
    public void visit(StatementList n) {
	ReturnValue rv = newReturnValue();
	Enumeration e = n.f0.elements();
	boolean goOn = e.hasMoreElements();
	lastStatementInStatementList = false;
	while (goOn) {
	    Node n2 = (Node)e.nextElement();
	    goOn = e.hasMoreElements();
	    if (!goOn) {
		lastStatementInStatementList = true;
	    }
	    n2.accept(this); rv.add (this.rv);
	    if (goOn) {
		nli(rv);
	    }
	}
	lastStatementInStatementList = false;
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ( &lt;IF&gt; &lt;LEFTPAREN&gt; Expression() &lt;RIGHTPAREN&gt; Statement() [ &lt;ELSE&gt; Statement() ] | &lt;SWITCH&gt; &lt;LEFTPAREN&gt; Expression() &lt;RIGHTPAREN&gt; Statement() )
     * </PRE>
     */
    public void visit(SelectionStatement n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ( &lt;WHILE&gt; &lt;LEFTPAREN&gt; Expression() &lt;RIGHTPAREN&gt; Statement() | &lt;DO&gt; Statement() &lt;WHILE&gt; &lt;LEFTPAREN&gt; Expression() &lt;RIGHTPAREN&gt; &lt;SEMICOLON&gt; | &lt;FOR&gt; &lt;LEFTPAREN&gt; [ Expression() ] &lt;SEMICOLON&gt; [ Expression() ] &lt;SEMICOLON&gt; [ Expression() ] &lt;RIGHTPAREN&gt; Statement() )
     * </PRE>
     */
    public void visit(IterationStatement n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ( &lt;GOTO&gt; &lt;IDENTIFIER&gt; &lt;SEMICOLON&gt; | &lt;CONTINUE&gt; &lt;SEMICOLON&gt; | &lt;BREAK&gt; &lt;SEMICOLON&gt; | &lt;RETURN&gt; [ Expression() ] &lt;SEMICOLON&gt; )
     * </PRE>
     */
    public void visit(JumpStatement n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> AssignmentExpression()
     * f1 -> ( &lt;COMMA&gt; AssignmentExpression() )*
     * </PRE>
     */
    public void visit(Expression n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> UnaryExpression() AssignmentOperator() AssignmentExpression()
     *       | ConditionalExpression()
     * </PRE>
     */
    public void visit(AssignmentExpression n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ( &lt;OP_ASS&gt; | &lt;OP_MULASS&gt; | &lt;OP_DIVASS&gt; | &lt;OP_MODASS&gt; | &lt;OP_ADDASS&gt; | &lt;OP_SUBASS&gt; | &lt;OP_SLASS&gt; | &lt;OP_SRASS&gt; | &lt;OP_ANDASS&gt; | &lt;OP_XORASS&gt; | &lt;OP_ORASS&gt; )
     * </PRE>
     */
    public void visit(AssignmentOperator n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> LogicalORExpression()
     * f1 -> [ &lt;QUESTION&gt; Expression() &lt;COLON&gt; ConditionalExpression() ]
     * </PRE>
     */
    public void visit(ConditionalExpression n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ConditionalExpression()
     * </PRE>
     */
    public void visit(ConstantExpression n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> LogicalANDExpression()
     * f1 -> [ &lt;OP_OR&gt; LogicalORExpression() ]
     * </PRE>
     */
    public void visit(LogicalORExpression n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> InclusiveORExpression()
     * f1 -> [ &lt;OP_AND&gt; LogicalANDExpression() ]
     * </PRE>
     */
    public void visit(LogicalANDExpression n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ExclusiveORExpression()
     * f1 -> [ &lt;OP_BITOR&gt; InclusiveORExpression() ]
     * </PRE>
     */
    public void visit(InclusiveORExpression n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ANDExpression()
     * f1 -> [ &lt;OP_BITXOR&gt; ExclusiveORExpression() ]
     * </PRE>
     */
    public void visit(ExclusiveORExpression n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> EqualityExpression()
     * f1 -> [ &lt;OP_BITAND&gt; ANDExpression() ]
     * </PRE>
     */
    public void visit(ANDExpression n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> RelationalExpression()
     * f1 -> [ ( &lt;OP_EQ&gt; | &lt;OP_NEQ&gt; ) EqualityExpression() ]
     * </PRE>
     */
    public void visit(EqualityExpression n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ShiftExpression()
     * f1 -> [ ( &lt;OP_LT&gt; | &lt;OP_GT&gt; | &lt;OP_LE&gt; | &lt;OP_GE&gt; ) RelationalExpression() ]
     * </PRE>
     */
    public void visit(RelationalExpression n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> AdditiveExpression()
     * f1 -> [ ( &lt;OP_SL&gt; | &lt;OP_SR&gt; ) ShiftExpression() ]
     * </PRE>
     */
    public void visit(ShiftExpression n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> MultiplicativeExpression()
     * f1 -> [ ( &lt;OP_ADD&gt; | &lt;OP_SUB&gt; ) AdditiveExpression() ]
     * </PRE>
     */
    public void visit(AdditiveExpression n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> CastExpression()
     * f1 -> [ ( &lt;OP_MUL&gt; | &lt;OP_DIV&gt; | &lt;OP_MOD&gt; ) MultiplicativeExpression() ]
     * </PRE>
     */
    public void visit(MultiplicativeExpression n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ( &lt;LEFTPAREN&gt; TypeName() &lt;RIGHTPAREN&gt; CastExpression() | UnaryExpression() )
     * </PRE>
     */
    public void visit(CastExpression n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ( PostfixExpression() | &lt;OP_INCR&gt; UnaryExpression() | &lt;OP_DECR&gt; UnaryExpression() | UnaryOperator() CastExpression() | &lt;SIZEOF&gt; ( UnaryExpression() | &lt;LEFTPAREN&gt; TypeName() &lt;RIGHTPAREN&gt; ) )
     * </PRE>
     */
    public void visit(UnaryExpression n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ( &lt;OP_BITAND&gt; | &lt;OP_MUL&gt; | &lt;OP_ADD&gt; | &lt;OP_SUB&gt; | &lt;OP_BITNOT&gt; | &lt;OP_NOT&gt; )
     * </PRE>
     */
    public void visit(UnaryOperator n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> PrimaryExpression()
     * f1 -> ( &lt;LEFTBRACKET&gt; Expression() &lt;RIGHTBRACKET&gt; | &lt;LEFTPAREN&gt; [ ArgumentExpressionList() ] &lt;RIGHTPAREN&gt; | &lt;DOT&gt; &lt;IDENTIFIER&gt; | &lt;OP_DEREF&gt; &lt;IDENTIFIER&gt; | &lt;OP_INCR&gt; | &lt;OP_DECR&gt; )*
     * </PRE>
     */
    public void visit(PostfixExpression n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> ( &lt;IDENTIFIER&gt; | Constant() | &lt;LEFTPAREN&gt; Expression() &lt;RIGHTPAREN&gt; )
     * </PRE>
     */
    public void visit(PrimaryExpression n) {
	ReturnValue rv = newReturnValue();

	if (n.f0.which == 0) { // it's an identifier
	    rv.add(replacedNameForIdentifier(n.f0));
	} else {
	    n.f0.accept(this); rv.add(this.rv);
	}

    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> AssignmentExpression()
     * f1 -> ( &lt;COMMA&gt; AssignmentExpression() )*
     * </PRE>
     */
    public void visit(ArgumentExpressionList n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
	n.f1.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }

    /**
     * <PRE>
     * f0 -> &lt;INTEGER_LITERAL&gt;
     *       | &lt;FLOATING_POINT_LITERAL&gt;
     *       | &lt;CHARACTER_LITERAL&gt;
     *       | &lt;STRING_LITERAL&gt;
     * </PRE>
     */
    public void visit(Constant n) {
	ReturnValue rv = newReturnValue();
	n.f0.accept(this); rv.add(this.rv);
    	this.rv = rv;
    }
}



